import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Represents a 6-sided box.
 * <p>
 * A {@code Box} is defined by a single point that represents the origin of the box
 * and three vectors defining the length, width, and height of the box.
 * </p>
 */
public class Box extends Shape
{
	/**
	 * Constructs a unit cube at the origin.
	 */
	public Box()
	{
		setOrigin( new Point3d(0,0,0) );
		setLength( new Vector3d(1,0,0) );
		setWidth( new Vector3d(0,1,0) );
		setHeight( new Vector3d(0,0,1) );
		constructBoxFaces();
	}

	/**
	 * Performs a deep copy of the specified {@code Box} object.
	 * 
	 * @param box
	 * 			The {@code Box} object to copy.
	 */
	protected Box( Box box )
	{
		super( box );
		setOrigin( (Point3d)box.getOrigin().clone() );
		setLength( (Vector3d)box.getLength().clone() );
		setWidth( (Vector3d)box.getWidth().clone() );
		setHeight( (Vector3d)box.getHeight().clone() );
		constructBoxFaces();
	}
	
	/**
	 * Constructs a box with the specified parameters.
	 * 
	 * @param origin
	 *            The origin point.
	 * @param length
	 *            The length vector.
	 * @param width
	 *            The width vector.
	 * @param height
	 *            The height vector.
	 * @param surfaceProperties
	 *            The surface properties of the box.
	 */
	public Box( Point3d origin, Vector3d length, Vector3d width,
			Vector3d height, SurfaceProperties surfaceProperties )
	{
		super( surfaceProperties );
		this.setOrigin( origin );
		this.setLength( length );
		this.setWidth( width );
		this.setHeight( height );
		constructBoxFaces();
	}

	@Override
	public Object clone()
	{
		return new Box( this ); 
	}

	public double nearestIntersectionTime( Ray3d ray )
	{
		double tMin = 0;
		double tMax = Double.POSITIVE_INFINITY;

		Vector3d rayToPlane = new Vector3d();
		for( int x = 0; x < 6; ++x )
		{
			Plane plane = faces[x];
			rayToPlane.sub( ray.getOrigin(), plane.getPoint() );
			double rtpDotN = rayToPlane.dot( plane.getNormal() );	    
			double dDotN = ray.getDirection().dot( plane.getNormal() );
			double intersection = 0;

			if( dDotN < 0 )
			{
				intersection = -rtpDotN / dDotN;
				if( intersection > tMin )
					tMin = intersection;
			}
			else if( dDotN > 0 )
			{
				intersection = -rtpDotN / dDotN;
				if( intersection < tMax )
					tMax = intersection;
			}
			else if( rtpDotN > 0 )
				tMin = tMax + 1;

			if( tMin > tMax )
				return Intersection.NoIntersection;
		}

		if( tMin > Scene.EPSILON )
			return tMin;
		else if( tMax > Scene.EPSILON )
			return tMax;
		else
			return Intersection.NoIntersection;

	}

	public Vector3d surfaceNormal( Point3d point )
	{
		// Check each plane to see if it contains the point
		// and return that planes normal if it does.
		for( Plane plane : faces )
		{
			if( plane.contains( point ) )
				return plane.getNormal();
		}

		// The point isn't contained in any of the boxes' planes.
		return null;
	}

	/**
	 * Constructs the faces of the box using the origin, length, width, and
	 * height fields of the {@code Box} instance.
	 */
	protected void constructBoxFaces()
	{
		faces = new Plane[6];

		// Create three planes from the box axis.
		faces[0] = new Plane( getOrigin(), getLength(), getHeight() );
		faces[1] = new Plane( getOrigin(), getHeight(), getWidth() );
		faces[2] = new Plane( getOrigin(), getWidth(), getLength() );

		// Create the three planes that are parallel 
		// to the planes created above.
		Point3d point = new Point3d( getOrigin() );
		
		point.add( getOrigin(), getWidth() );
		faces[3] = new Plane( point, getHeight(), getLength() );
		point.add( getOrigin(), getLength() );
		faces[4] = new Plane( point, getWidth(), getHeight() );
		point.add( getOrigin(), getHeight() );
		faces[5] = new Plane( point, getLength(), getWidth() );
	}

	public void load( DataParser dataParser )
	{
		setOrigin( dataParser.getPoint3d() );
		setLength( dataParser.getVector3d() );
		setWidth( dataParser.getVector3d() );
		setHeight( dataParser.getVector3d() );	
		super.load( dataParser );
		constructBoxFaces();
	}

	/**
	 * Sets the origin point.
	 * 
	 * @param origin
	 *            The origin to set.
	 */
	public void setOrigin( Point3d origin )
	{
		this.origin = origin;
	}

	/**
	 * Returns the origin point.
	 * 
	 * @return The origin.
	 */
	public Point3d getOrigin()
	{
		return origin;
	}

	/**
	 * Sets the length vector.
	 * 
	 * @param length
	 *            The length to set.
	 */
	public void setLength( Vector3d length )
	{
		this.length = length;
	}

	/**
	 * Returns the length vector.
	 * 
	 * @return The length.
	 */
	public Vector3d getLength()
	{
		return length;
	}

	/**
	 * Sets the width vector.
	 * 
	 * @param width
	 *            The width to set.
	 */
	public void setWidth( Vector3d width )
	{
		this.width = width;
	}

	/**
	 * Returns the width vector.
	 * 
	 * @return The width.
	 */
	public Vector3d getWidth()
	{
		return width;
	}

	/**
	 * Sets the height vector.
	 * 
	 * @param height
	 *            The height to set.
	 */
	public void setHeight( Vector3d height )
	{
		this.height = height;
	}

	/**
	 * Returns the height vector.
	 * 
	 * @return The height.
	 */
	public Vector3d getHeight()
	{
		return height;
	}

	private Point3d origin;
	private Vector3d length;
	private Vector3d width;
	private Vector3d height;
	private Plane[] faces;
}