import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * A simple camera class for specifying the view into the scene.
 * <p>
 * Our 'camera' model is as follows. The center of the view plane is located at
 * c = (cx, cy, cz). The view window (whose center is also c) is defined by two
 * vectors u = (ux, uy, uz) and v = (vx, vy, vz) in the view plane:
 * <p>
 * <center> view window = {{@literal c + au + bv | -1 <= a <= 1,-1 <= b <= 1}}
 * (the vectors u and v are assumed to be orthogonal). </center>
 * </p>
 * <p>
 * The view reference point (eye) is located at c + e, with e = (ex, ey, ez).
 * <p>
 * Some things to note about our camera model:
 * <ul>
 * <li>The eye vector e is not necessarily orthogonal to the view plane!</li>
 * <li>The width of the view window is 2|u|, and the height is 2|v|.</li>
 * <li>The normal to the view window/plane is u x v.</li>
 * </ul>
 * }
 */
public class Camera implements Persistent, CloneableEx
{
	/**
	 * Constructs a camera with a view window centered at the origin on the XY
	 * plane, with a width (x-axis) and height (y-axis) of 1, and the eye point
	 * at (0,0,1) looking down the negative z-axis.
	 */
	public Camera()
	{
		setViewCenter( new Point3d(0,0,0) );
		setWidthVector( new Vector3d(0.5,0,0) );
		setHeightVector( new Vector3d(0,0.5,0) );
		setEyeVector( new Vector3d(0,0,1) );
	}

	/**
	 * Performs a deep copy of the specified {@code Camera} object.
	 * 
	 * @param camera
	 * 			The {@code Camera} object to copy.
	 */
	protected Camera( Camera camera )
	{
		setViewCenter( (Point3d)camera.getViewCenter().clone() );
		setWidthVector( (Vector3d)camera.getWidthVector().clone() );
		setHeightVector( (Vector3d )camera.getHeightVector().clone() );
		setEyeVector( (Vector3d)camera.getEyeVector().clone() );
	}
	
	/**
	 * Constructs a camera with the specified parameters.
	 * 
	 * @param viewCenter
	 *            The center of the view window.
	 * @param widthVector
	 *            The vector from the {@code viewCenter} to positive
	 *            horizontal edge of the view window.
	 * @param heightVector
	 *            The vector from the {@code viewCenter} to positive
	 *            vertical edge of the view window.
	 * @param eyeVector
	 *            The vector from the {@code viewCenter} to the eye point.
	 */
	public Camera( Point3d viewCenter, Vector3d widthVector, Vector3d heightVector, Vector3d eyeVector )
	{
		setViewCenter( viewCenter );
		setWidthVector( widthVector );
		setHeightVector( heightVector );
		setEyeVector( eyeVector );
	}

	@Override
	public Object clone()
	{
		return new Camera( this );
	}

	public void load( DataParser dataParser )
	{
		setViewCenter( dataParser.getPoint3d() );
		setWidthVector( dataParser.getVector3d() );
		setHeightVector( dataParser.getVector3d() );
		setEyeVector( dataParser.getVector3d() );
	}

	/**
	 * Sets the viewCenter.
	 * 
	 * @param viewCenter
	 *            The viewCenter to set.
	 */
	public void setViewCenter( Point3d viewCenter )
	{
		this.viewCenter = viewCenter;
	}

	/**
	 * Returns the viewCenter.
	 * 
	 * @return The viewCenter.
	 */
	public Point3d getViewCenter()
	{
		return viewCenter;
	}

	/**
	 * Sets the widthVector.
	 * 
	 * @param widthVector
	 *            The widthVector to set.
	 */
	public void setWidthVector( Vector3d widthVector )
	{
		this.widthVector = widthVector;
	}

	/**
	 * Returns the widthVector.
	 * 
	 * @return The widthVector.
	 */
	public Vector3d getWidthVector()
	{
		return widthVector;
	}

	/**
	 * Sets the heightVector.
	 * 
	 * @param heightVector
	 *            The heightVector to set.
	 */
	public void setHeightVector( Vector3d heightVector )
	{
		this.heightVector = heightVector;
	}

	/**
	 * Returns the heightVector.
	 * 
	 * @return The heightVector.
	 */
	public Vector3d getHeightVector()
	{
		return heightVector;
	}

	/**
	 * Sets the eyeVector.
	 * 
	 * @param eyeVector
	 *            The eyeVector to set.
	 */
	public void setEyeVector( Vector3d eyeVector )
	{
		this.eyeVector = eyeVector;
	}

	/**
	 * Returns the eyeVector.
	 * 
	 * @return The eyeVector.
	 */
	public Vector3d getEyeVector()
	{
		return eyeVector;
	}

	private Point3d viewCenter;
	private Vector3d widthVector;
	private Vector3d heightVector;
	private Vector3d eyeVector;
}