import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.regex.Pattern;

import javax.vecmath.Color3f;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Parses data from a string or text file. The input data format is not
 * documented here yet. Sorry!
 */
public class DataParser
{
	/**
	 * Constructs a data parser with default state.
	 */
	public DataParser()
	{
		tokenIndex = 0;
		this.tokens = null;
	}

	/**
	 * Parses the input sequence using the specified regular expression, and
	 * sets the data parser to the first token.
	 * 
	 * @param input
	 *            The character data to parse.
	 * @param regularExpression
	 *            The regular expression used to parse the character data.
	 * @return Returns {code true} if successful, {code false}
	 *         otherwise.
	 */
	public boolean parse( CharSequence input, String regularExpression )
	{
		if( input == null || regularExpression == null )
			return false;

		Pattern pattern = Pattern.compile( regularExpression );
		tokens = pattern.split( input );
		tokenIndex = 0;

		return true;
	}

	/**
	 * Parses data from a string.
	 * 
	 * @param dataString
	 *            The character data to parse.
	 * @param regularExpression
	 *            The regular expression to parse the character data with.
	 * @return Returns {code true} if successful, {code false}
	 *         otherwise.
	 */
	public boolean parseString( String dataString, String regularExpression )
	{
		return parse( dataString, regularExpression );
	}

	/**
	 * Parses data from a file.
	 * 
	 * @param filename
	 *            The name of the file with character data to parse.
	 * @param regularExpression
	 *            The regular expression to parse the file with.
	 * @return Returns {code true} if successful, {code false}
	 *         otherwise.
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public boolean parseFile( String filename, String regularExpression ) throws FileNotFoundException, IOException
	{
//		if( filename == null )
//			return false;

		StringBuilder buffer = createStringFromFile( filename );

		if( buffer != null )
			return parse( buffer, regularExpression );

		return false;
	}

	/**
	 * Creates a single character string from a file.
	 * 
	 * @param filename
	 *            The name of the file to create a string out of.
	 * @return The string containing the file contents.
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public StringBuilder createStringFromFile( String filename ) throws FileNotFoundException, IOException
	{
		StringBuilder buffer = null;

		// Create a File
		File inFile = new File( filename );
		int approximateLength = (int)inFile.length();

		// Create a Character Input Stream
		FileReader inStream = new FileReader( inFile );

		// Filter the Input Stream - buffers characters for efficiency
		BufferedReader in = new BufferedReader( inStream );

		buffer = new StringBuilder( approximateLength );
		String line;
		while( (line = in.readLine()) != null )
		{
			if(  line.length() > 0 && line.charAt(0) != '#' )
			{
				// The readLine() method discards line feeds and carriage returns 
				// which may cause improper concatenation of two lines if the 
				// whitespace is not appended first.
				buffer.append( ' ' );
				buffer.append( line );
			}
		}


		return buffer;
	}

	/**
	 * Checks whether there are any tokens remaining.
	 * 
	 * @return Returns {code true} if there are more tokens, otherwise
	 *         {code false}.
	 */
	public boolean tokensRemaining()
	{
		boolean isValid = false;
		if( tokens != null )
			isValid = tokenIndex < tokens.length;

		return isValid;
	}

	/**
	 * Gets an integer and advances the token stream.
	 * 
	 * @return An integer value.
	 */
	public int getInteger()
	{
		return Integer.parseInt( tokens[ tokenIndex++ ] );
	}

	/**
	 * Gets a double and advances the token stream.
	 * 
	 * @return A double value.
	 */
	public double getDouble()
	{
		return Double.parseDouble( tokens[ tokenIndex++ ] ); 
	}

	/**
	 * Gets a float and advances the token stream.
	 * 
	 * @return A float value.
	 */
	public float getFloat()
	{ 
		return Float.parseFloat( tokens[ tokenIndex++ ] ); 
	}

	/**
	 * Gets a string and advances the token stream.
	 * 
	 * @return A string value.
	 */
	public String getString() 
	{ 
		return tokens[ tokenIndex++ ]; 
	}

	/**
	 * Gets a vector and advances the token stream.
	 * 
	 * @return A vector value.
	 */
	public Vector3d getVector3d()
	{
		return new Vector3d( getDouble(), getDouble(), getDouble() );
	}
	
	/**
	 * Gets a point and advances the token stream.
	 * 
	 * @return A point value.
	 */
	public Point3d getPoint3d()
	{
		return new Point3d( getDouble(), getDouble(), getDouble() );
	}

	/**
	 * Gets a color and advances the token stream.
	 * 
	 * @return A color value.
	 */
	public Color3f getColor3f()
	{
		return new Color3f( getFloat(), getFloat(), getFloat() );
	}

	private String[] tokens;
	private int tokenIndex;
}