import javax.vecmath.*;

/**
 * Represents an ellipsoid
 * 
 * An {code Ellipsoid} is specified by an origin point and a
 * vector for each of it's semi-axes.
 */
public class Ellipsoid extends Shape
{
	/**
	 * Constructs an {@code Ellipsoid} at the origin, with unit length semi-axes (a sphere).
	 */
	public Ellipsoid()
	{
		setOrigin( new Point3d() );
		setSemiAxisX( new Vector3d(1.0, 0.0, 0.0) );
		setSemiAxisY( new Vector3d(0.0, 1.0, 0.0) );
		setSemiAxisZ( new Vector3d(0.0, 0.0, 1.0) );
		calculateInverseTransform();
	}
	
	/**
	 * Performs a deep copy of the specified {@code Ellipsoid} object.
	 * 
	 * @param ellipsoid
	 * 			The {@code Ellipsoid} object to copy.
	 */
	protected Ellipsoid( Ellipsoid ellipsoid )
	{
		super( ellipsoid );
		setOrigin( (Point3d)ellipsoid.getOrigin().clone() );
		setSemiAxisX( (Vector3d)ellipsoid.getSemiAxisX().clone() );
		setSemiAxisY( (Vector3d)ellipsoid.getSemiAxisY().clone() );
		setSemiAxisZ( (Vector3d)ellipsoid.getSemiAxisZ().clone() );
		setTransform( (Matrix3d)ellipsoid.getTransform().clone() );
	}
	
	/**
	 * Constructs an {@code Ellipsoid} with the specified parameters.
	 * 
	 * @param origin
	 * 			The origin of the ellipsoid.
	 * @param semiaxisA
	 * 			The semi-axis along the X axis in object space.
	 * @param semiaxisB
	 * 			The semi-axis along the Y axis in object space.
	 * @param semiaxisC
	 * 			The semi-axis along the Z axis in object space.
	 * @param sp
	 * 			The surface properties of the ellipsoid.
	 */
	public Ellipsoid( Point3d origin, Vector3d semiaxisA, Vector3d semiaxisB, Vector3d semiaxisC, SurfaceProperties sp )
	{
		super( sp );
		this.setOrigin( origin );
		this.setSemiAxisX( semiaxisA );
		this.setSemiAxisY( semiaxisB );
		this.setSemiAxisZ( semiaxisC );
		calculateInverseTransform();
	}

	@Override
	public Object clone()
	{
		return new Ellipsoid( this );
	}

	public double nearestIntersectionTime( Ray3d ray )
	{
		Vector3d transfDirection = new Vector3d( ray.getDirection() );
		getTransform().transform( transfDirection );
		double A = transfDirection.lengthSquared();
		Vector3d transfCenters = new Vector3d();
		transfCenters.sub( ray.getOrigin(), this.getOrigin() );
		getTransform().transform( transfCenters );
		double B = 2 * transfCenters.dot( transfDirection );
		double C = transfCenters.lengthSquared() - 1;

		double discriminant = ( B * B ) - ( 4 * A * C );
		if( discriminant < 0.0 )
			return Intersection.NoIntersection;
		else
		{
			double sqrtdiscriminant = Math.sqrt( discriminant );
			double timeExit = ( -B + sqrtdiscriminant ) / ( 2 * A );
			double timeEnter = ( -B - sqrtdiscriminant ) / ( 2 * A );
			if( timeExit < Scene.EPSILON )
				return Intersection.NoIntersection;
			else if( timeEnter < Scene.EPSILON )
				return timeExit;
			else
				return timeEnter;
		}
	}

	public Vector3d surfaceNormal( Point3d point )
	{
		Vector3d normal = new Vector3d();
		normal.sub( point, this.getOrigin() );
		getTransform().transform( normal );
		Matrix3d transformTranspose = new Matrix3d();
		transformTranspose.transpose( getTransform() );
		transformTranspose.transform( normal );
		return normal;	
	}

	public void load( DataParser dataParser )
	{
		setOrigin( dataParser.getPoint3d() );
		setSemiAxisX( dataParser.getVector3d() );
		setSemiAxisY( dataParser.getVector3d() );
		setSemiAxisZ( dataParser.getVector3d() );	
		super.load( dataParser );
		calculateInverseTransform();
	}

	/**
	 * Calculates the matrix that transforms the ellipsoid from world space
	 * to a sphere centered on the origin.
	 */
	private void calculateInverseTransform()
	{
		setTransform( new Matrix3d(
				getSemiAxisX().x, getSemiAxisY().x, getSemiAxisZ().x,
				getSemiAxisX().y, getSemiAxisY().y, getSemiAxisZ().y,
				getSemiAxisX().z, getSemiAxisY().z, getSemiAxisZ().z
				) );
		
		getTransform().invert();
	}

	/**
	 * Sets the origin of the ellipsoid.
	 * 
	 * @param origin
	 * 			The new origin point.
	 */
	public void setOrigin( Point3d origin )
	{
		this.origin = origin;
	}

	/**
	 * Returns the origin of the ellipsoid.
	 * 
	 * @return
	 * 			The origin point.
	 */
	public Point3d getOrigin()
	{
		return origin;
	}

	/**
	 * Sets the semi-axis corresponding to the X axis in object space.
	 * 
	 * @param semiAxisX
	 * 			The new X semi-axis.
	 */
	public void setSemiAxisX( Vector3d semiAxisX )
	{
		this.semiAxisX = semiAxisX;
	}

	/**
	 * Returns the semi-axis corresponding to the X axis in object space.
	 * 
	 * @return
	 * 			The X semi-axis.
	 */
	public Vector3d getSemiAxisX()
	{
		return semiAxisX;
	}

	/**
	 * Sets the semi-axis corresponding to the Y axis in object space.
	 * 
	 * @param semiAxisY
	 * 			The new Y semi-axis.
	 */
	public void setSemiAxisY( Vector3d semiAxisY )
	{
		this.semiAxisY = semiAxisY;
	}

	/**
	 * Returns the semi-axis corresponding to the Y axis in object space.
	 * 
	 * @return
	 * 			The Y semi-axis.
	 */
	public Vector3d getSemiAxisY()
	{
		return semiAxisY;
	}

	/**
	 * Sets the semi-axis corresponding to the Z axis in object space.
	 * 
	 * @param semiAxisZ
	 * 			The new Z semi-axis.
	 */
	public void setSemiAxisZ( Vector3d semiAxisZ )
	{
		this.semiAxisZ = semiAxisZ;
	}

	/**
	 * Returns the semi-axis corresponding to the Z axis in object space.
	 * 
	 * @return
	 * 			The Z semi-axis.
	 */
	public Vector3d getSemiAxisZ()
	{
		return semiAxisZ;
	}

	/**
	 * Sets the matrix that transforms the ellipsoid from world space
	 * to a sphere centered on the origin.
	 * 
	 * @param transform
	 * 			The transform to set.
	 */
	private void setTransform( Matrix3d transform )
	{
		this.transform = transform;
	}

	/**
	 * Returns the matrix that transforms the ellipsoid from world space
	 * to a sphere centered on the origin.
	 * 
	 * @return
	 * 			The transformation matrix.
	 */
	public Matrix3d getTransform()
	{
		return transform;
	}

	private Point3d origin;
	private Vector3d semiAxisX;
	private Vector3d semiAxisY;
	private Vector3d semiAxisZ;
	private Matrix3d transform;
}