import javax.vecmath.Color3f;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * A class representing a light.
 * 
 * A light is specified by it's position, color, and radius.
 */
public class Light implements Persistent, CloneableEx
{
	public Light()
	{
		setPosition( new Point3d() );
		setColor( new Color3f() );
		setRadius( 0.0 );
		width = new Vector3d();
		height = new Vector3d();
		xSamples = 0;
		ySamples = 0;
	}

	/**
	 * Performs a deep copy of the specified {@code Light} object.
	 * 
	 * @param light
	 * 			The {@code Light} object to copy.
	 */
	public Light( Light light )
	{
		setPosition( (Point3d)light.getPosition().clone() );
		setColor( (Color3f)light.getColor().clone() );
		setRadius( light.getRadius() );
		width = (Vector3d)light.width.clone();
		height = (Vector3d)light.height.clone();
		xSamples = light.xSamples;
		ySamples = light.ySamples;
	}
	
	/**
	 * Constructs a new point light with the specified parameters.
	 * 
	 * @param position
	 * 			The position of the light.
	 * @param color
	 * 			The color of the light.
	 * @param radius
	 * 			The radius the light extends. This parameter is unused.
	 */
	public Light( Point3d position, Color3f color, double radius )
	{
		this.setPosition( position );
		this.setColor( color );
		this.setRadius( radius );
	}

	/**
	 * Constructs an extended light. This is unimplemented.
	 * 
	 * @param position
	 * 			The position of the light.
	 * @param color
	 * 			The color of the light.
	 * @param radius
	 * 			The radius the light extends. This parameter is unused.
	 * @param width
	 * 			The width of the light.
	 * @param height
	 * 			The height of the light.
	 * @param xSamples
	 * 			The number of samples taken along the width axis.
	 * @param ySamples
	 * 			The number of samples taken along the height axis.
	 */
	@SuppressWarnings("unused")
	private Light( Point3d position, Color3f color, double radius, Vector3d width, Vector3d height, int xSamples, int ySamples )
	{
		this.setPosition( position );
		this.setColor( color );
		this.setRadius( radius );
		this.width = width;
		this.height = height;
		this.xSamples = xSamples;
		this.ySamples = ySamples;
	}

	@Override
	public Object clone()
	{
		return new Light( this );
	}

	public void load( DataParser dataParser )
	{
		setPosition( dataParser.getPoint3d() );
		setColor( dataParser.getColor3f() );
		setRadius( dataParser.getDouble() );
		
		// These are unused for the time being.
		width = new Vector3d(1,0,0);//dataParser.getVector3d();
		height = new Vector3d(0,1,0);//dataParser.getVector3d();
		xSamples = 1;//dataParser.getInteger();
		ySamples = 1;//dataParser.getInteger();
	}

	/**
	 * Sets the position of the light.
	 * @param position The position to set.
	 */
	public void setPosition( Point3d position )
	{
		this.position = position;
	}

	/**
	 * Returns the position of the light.
	 * @return The position of the light.
	 */
	public Point3d getPosition()
	{
		return position;
	}

	/**
	 * Sets the color of the light.
	 * @param color The color to set.
	 */
	public void setColor( Color3f color )
	{
		this.color = color;
	}

	/**
	 * Returns the color of the light.
	 * @return The color of the light.
	 */
	public Color3f getColor()
	{
		return color;
	}

	/**
	 * Sets the radius of the light.
	 * @param radius The radius to set.
	 */
	public void setRadius( double radius )
	{
		this.radius = radius;
	}

	/**
	 * Returns the radius of the light.
	 * @return The radius of the light.
	 */
	public double getRadius()
	{
		return radius;
	}

	private Point3d position;
	private Color3f color;
	private double radius;
	
	private Vector3d width;
	private Vector3d height;
	private int xSamples;
	private int ySamples;
}