import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;

/**
 * A component adapter to ensure a component does not get resized below the 
 * specified dimensions.
 */
public class MinimumFrameSizeComponent extends ComponentAdapter
{
	/**
	 * Constructs a {@code MinimumFrameSizeComponent} that restricts a 
	 * component to the {@code minimumWidth} and {@code minimumHeight}
	 * in pixels.
	 * @param minimumWidth
	 * 			The minimum width in pixels.
	 * @param minimumHeight
	 * 			The minimum height in pixels.
	 */
	public MinimumFrameSizeComponent( int minimumWidth, int minimumHeight )
	{
		minWidth = minimumWidth;
		minHeight = minimumHeight;
	}

	/**
	 * Called when a component is resized, and prevents the component from
	 * becoming smaller than {@code minimumWidth} and {@code minimumHeight}.
	 */
	public void componentResized( ComponentEvent e )
	{
		int width = e.getComponent().getWidth();
		int height = e.getComponent().getHeight();

		boolean resize = false;

		// Clamp the new width to the minimum width.
		if( width < minWidth )
		{
			resize = true;
			width = minWidth;
		}

		// Clamp the new height to the minimum height.
		if( height < minHeight )
		{
			resize = true;
			height = minHeight;
		}

		if( resize )
			e.getComponent().setSize( width, height );
	}

	private int minWidth;
	private int minHeight;
}