import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * A class representing a plane.
 * 
 * A plane is specified by a point and a normal.
 */
public class Plane extends Shape
{
	public Plane()
	{
		setPoint( new Point3d() );
		setNormal( new Vector3d() );
	}

	/**
	 * Performs a deep copy of the specified {@code Plane} object.
	 * 
	 * @param plane
	 * 			The {@code Plane} object to copy.
	 */
	protected Plane( Plane plane )
	{
		super( plane );
		setPoint( (Point3d)plane.getPoint().clone() );
		setNormal( (Vector3d)plane.getNormal().clone() );
	}
	
	/**
	 * Constructs a plane from a point and the plane's normal.
	 * 
	 * @param point
	 * 			A point on the plane.
	 * @param normal
	 * 			The plane's normal.
	 */
	public Plane( Point3d point, Vector3d normal )
	{
		this.setPoint( point );
		this.setNormal( normal );
	}

	/**
	 * Constructs a plane from a point and two vectors laying in the plane. 
	 * 
	 * @param point
	 * 			A point on the plane.
	 * @param v1
	 * 			A vector on the plane.
	 * @param v2
	 * 			A vector on the plane.
	 */
	public Plane( Point3d point, Vector3d v1, Vector3d v2 )
	{
		this.setPoint( new Point3d() );
		this.setNormal( new Vector3d() );
		createPlaneFromVectors( point, v1, v2 );
	}

	@Override
	public Object clone()
	{
		return new Plane( this );
	}

	/**
	 * Constructs a plane from a point and two vectors laying in the plane. 
	 * 
	 * @param point
	 * 			A point on the plane.
	 * @param v1
	 * 			A vector on the plane.
	 * @param v2
	 * 			A vector on the plane.
	 */
	public void createPlaneFromVectors( Point3d point, Vector3d v1, Vector3d v2 )
	{
		this.getPoint().set( point );
		getNormal().cross( v1, v2 );
	}

	/**
	 * Determines if a point is contained on the plane.
	 * 
	 * @param point
	 * 			The point to test.
	 * @return
	 * 			{@code true} if the point is on the plane, otherwise {@code false}.
	 */
	public boolean contains( Point3d point )
	{
		Vector3d pointToPlane = new Vector3d();
		pointToPlane.sub( point, this.getPoint() );
		double nDotP = pointToPlane.dot( getNormal() );
		if( Math.abs( nDotP ) < Scene.EPSILON )
			return true;

		return false;
	}

	public double nearestIntersectionTime( Ray3d ray )
	{
		Vector3d v = new Vector3d();
		v.sub( ray.getOrigin(), getPoint() );
		double dDotN = ray.getDirection().dot( getNormal() );
		if( dDotN != 0 )
		{
			double intersection = -( v.dot( getNormal() ) / dDotN );
			if( intersection > Scene.EPSILON )
				return intersection;
		}        

		return Intersection.NoIntersection;
	}

	public Vector3d surfaceNormal( Point3d point )
	{
		return getNormal();
	}

	/**
	 * Sets the point on the plane.
	 * @param point The point to set.
	 */
	public void setPoint( Point3d point )
	{
		this.point = point;
	}

	/**
	 * Returns the point on the plane.
	 * @return The point on the plane.
	 */
	public Point3d getPoint()
	{
		return point;
	}

	/**
	 * Sets the plane's normal.
	 * @param normal The normal to set.
	 */
	public void setNormal( Vector3d normal )
	{
		this.normal = normal;
	}

	/**
	 * Returns the plane's  normal.
	 * @return The plane's normal.
	 */
	public Vector3d getNormal()
	{
		return normal;
	}

	private Point3d point;
	private Vector3d normal;    
}