import java.util.ArrayList;

import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;


/**
 * A class representing a polygon.
 * 
 * A polygon is specified by a list of vertices.
 * 
 * If the vertices are not coplanar methods called on this class are undefined.
 */
public class Polygon extends Shape
{
	public Polygon()
	{
		setVertices( new ArrayList<Point3d>() );
		setPlane( new Plane() );
	}
	
	/**
	 * Performs a deep copy of the specified {@code Polygon} object.
	 * 
	 * @param polygon
	 * 			The {@code Polygon} object to copy.
	 */
	protected Polygon( Polygon polygon )
	{
		super( polygon );
		setVertices( new ArrayList< Point3d >( polygon.getVertices().size() ) );
		for( Point3d p : polygon.getVertices() )
			getVertices().add( (Point3d)p.clone() );
		
		setPlane( (Plane)polygon.getPlane().clone() );
	}
	
	/**
	 * Constructs a polygon with the specified parameters.
	 * 
	 * The vertices should be coplanar.
	 * 
	 * @param vertices
	 * 			The vertices making up the polygon.
	 * @param sp
	 * 			The surface properties of the polygon.
	 */
	public Polygon( ArrayList<Point3d> vertices, SurfaceProperties sp )
	{
		super( sp );
		this.setVertices( new ArrayList<Point3d>( vertices ) );
		createPolygonPlane();	
	}

	@Override
	public Object clone()
	{
		return new Polygon( this );
	}

	public double nearestIntersectionTime( Ray3d ray )
	{
		// Find the ray/plane intersection time.
		double intersection = getPlane().nearestIntersectionTime( ray );
		if( intersection == Intersection.NoIntersection )
			return Intersection.NoIntersection;

		Point3d p = ray.interpolate( intersection );
		Point3d c = getVertices().get( 0 );
		Vector3d v = new Vector3d();
		Vector3d v1 = new Vector3d();
		Vector3d v2 = new Vector3d();
		v.sub( p, c );

		for( int x = 2; x < getVertices().size(); ++x )
		{
			v1.sub( getVertices().get( x - 1 ), c );
			v2.sub( getVertices().get( x ), c );
			double v1Dotv2 = v1.dot( v2 );
			double v1LengthSqr = v1.lengthSquared();
			double v2LengthSqr = v2.lengthSquared();
			double determinant = 1 / ( v1LengthSqr * v2LengthSqr - v1Dotv2 * v1Dotv2 );

			double M00 = v2LengthSqr * determinant;
			double M01 = -v1Dotv2 * determinant;
			double M10 = M01;
			double M11 = v1LengthSqr * determinant;

			double vDotv1 = v.dot( v1 );
			double vDotv2 = v.dot( v2 );
			double alpha = M00 * vDotv1 + M01 * vDotv2;
			double beta = M10 * vDotv1 + M11 * vDotv2;

			if( alpha >= 0 && beta >= 0 )
			{
				if( ( alpha + beta ) <= 1 )
					return intersection;
			}
		}

		return Intersection.NoIntersection;
	}

	public Vector3d surfaceNormal( Point3d point )
	{
		return getPlane().getNormal();
	}

	/**
	 * Creates the plane of the polygon using the first three vertices.
	 * 
	 * If there aren't at least three vertices, nothing is done.
	 */
	private void createPolygonPlane()
	{
		if( vertices.size() < 3 )
			return;
		
		Vector3d v1 = new Vector3d();
		Vector3d v2 = new Vector3d();
		v1.sub( getVertices().get(1), getVertices().get(0) );
		v2.sub( getVertices().get(2), getVertices().get(0) );
		getPlane().createPlaneFromVectors( getVertices().get(0), v1, v2 );
	}

	public void load( DataParser dataParser )
	{
		int vertexCount = dataParser.getInteger();
		getVertices().ensureCapacity( vertexCount );
		for( int i = 0; i < vertexCount; i++ )
			getVertices().add( dataParser.getPoint3d() );

		super.load( dataParser );
		createPolygonPlane();
	}

	/**
	 * Sets the vertices of the polygon.
	 * 
	 * @param vertices The vertices to set.
	 */
	public void setVertices( ArrayList<Point3d> vertices )
	{
		this.vertices = vertices;
		createPolygonPlane();
	}

	/**
	 * Returns the vertices of the polygon.
	 * @return The vertices of the polygon.
	 */
	public ArrayList<Point3d> getVertices()
	{
		return vertices;
	}

	/**
	 * Sets the plane of the polygon.
	 * @param plane The plane to set.
	 */
	private void setPlane( Plane plane )
	{
		this.plane = plane;
	}

	/**
	 * Returns the plane of the polygon.
	 * @return The plane of the polygon.
	 */
	public Plane getPlane()
	{
		return plane;
	}

	private ArrayList<Point3d> vertices;
	private Plane plane;
}