import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * A class for manipulating 3-dimensional rays.
 * 
 * A {@code Ray3d} is specified by an origin point and a direction vector.
 */
public class Ray3d
{
	/**
	 * Constructs a new {@code Ray3d} with an origin of (0, 0, 0) and a direction of (0,
	 * 0, 0).
	 */
	public Ray3d()
	{
		set( new Point3d(), new Vector3d() );
	}

	/**
	 * Constructs a new ray initialized to the specified origin and direction.
	 * 
	 * The direction is normalized.
	 * 
	 * @param origin
	 * 			The ray's origin.
	 * @param direction
	 * 			The ray's direction.
	 */
	public Ray3d( Point3d origin, Vector3d direction )
	{
		setNormalized( new Point3d( origin ), new Vector3d( direction ) );
	}

	/**
	 * Sets the origin and direction of the ray.
	 * 
	 * @param origin
	 * 			The ray's origin.
	 * @param direction
	 * 			The ray's direction.
	 */
	public void set( Point3d origin, Vector3d direction )
	{
		setOrigin( origin );
		setDirection( direction );
	}

	/**
	 * Sets the origin and direction of the ray, after normalizing the direction.
	 * 
	 * @param origin
	 * 			The ray's origin.
	 * @param direction
	 * 			The ray's direction.
	 */
	public void setNormalized( Point3d origin, Vector3d direction )
	{
		direction.normalize();
		
		setOrigin( origin );
		setDirection( direction );
	}

	/**
	 * Returns the point along the ray that is at the specified time from the
	 * origin.
	 * 
	 * @param time
	 * 			The time from the ray's origin.
	 * @return
	 * 			The point on the ray at {@code time}.
	 */
	public Point3d interpolate( double time )
	{
		Point3d point = new Point3d();
		point.scaleAdd( time, getDirection(), getOrigin() );
		return point;
	}

	/**
	 * Sets the origin of the ray.
	 * 
	 * @param origin
	 *            The ray's origin.
	 */
	public void setOrigin( Point3d origin )
	{
		this.origin = origin;
	}

	/**
	 * Returns the origin of the ray.
	 * 
	 * @return The ray's origin.
	 */
	public Point3d getOrigin()
	{
		return origin;
	}

	/**
	 * Sets the direction of the ray.
	 * 
	 * @param direction
	 *            The ray's direction.
	 */
	public void setDirection( Vector3d direction )
	{
		this.direction = direction;
	}

	/**
	 * Returns the direction of the ray.
	 * 
	 * @return The ray's direction.
	 */
	public Vector3d getDirection()
	{
		return direction;
	}

	private Point3d origin;
	private Vector3d direction;
}
