import java.awt.*;
import java.awt.image.*;
import java.util.ArrayList;
import javax.vecmath.*;

/**
 * A class for ray tracing a scene.
 */
public abstract class RayTracer
{
	public RayTracer( Scene scene )
	{
		this.scene = scene;
		this.imageBuffer = new BufferedImage( scene.getWidth(), scene.getHeight(), BufferedImage.TYPE_3BYTE_BGR );
		this.pixelsShaded = 0;
	}

	/**
	 * An abstract method for calculating the initial position the ray is cast through.
	 * 
	 * This point will be used as the base point to step through all the pixels
	 * in the viewport.
	 * 
	 * @param pixel
	 * 			The point that will contain the initial pixel position.
	 */
	protected abstract void calculateSamplingInfo( Point3d pixel );

	/**
	 * An abstract method to sample a pixel.
	 * 
	 * @param pixel
	 * 			The pixel to be sampled.
	 * @param widthStep
	 * 			The width of the pixel.
	 * @param heightStep
	 * 			The height of the pixel.
	 * @return
	 * 			The color sampled from the specified pixel.
	 */
	protected abstract Color3f samplePixel( Point3d pixel, Vector3d widthStep,
			Vector3d heightStep );

	/**
	 * Render the scene, updating after every row of pixels.
	 * 
	 * @param progressUpdater
	 * 			The object that will process progress updates.
	 */
	public void render( IProgressable progressUpdater )
	{
		if( imageBuffer == null || scene == null )
			return;

		// Calculate the step sizes and sampling info.
		Dimension viewportDims = scene.getViewportDimensions();
		WritableRaster raster = imageBuffer.getRaster();

		Vector3d widthStep = new Vector3d();
		Vector3d heightStep = new Vector3d();
		widthStep.scale( 2.0, scene.getCamera().getWidthVector() );
		heightStep.scale( 2.0, scene.getCamera().getHeightVector() );
		widthStep.scale( 1.0 / viewportDims.width );
		heightStep.scale( 1.0 / viewportDims.height );

		Point3d pixel = new Point3d();
		calculateSamplingInfo( pixel );

		// Step through all pixels in the viewport and cast a ray.
		pixelsShaded = 0;
		Point3d currentPixel = new Point3d();
		for( int y = 0; y < viewportDims.height; y ++ )
		{
			currentPixel.set( pixel );
			for( int x = 0; x < viewportDims.width; x++ )
			{
				// Abort render if the thread was interrupted.
				if( Thread.currentThread().isInterrupted() )
				    break;
				
				// Sample the pixel, and save it to the raster.
				Color3f color = samplePixel( currentPixel, widthStep, heightStep );
				color.clampMax( 1.0f );
				color.clampMin( 0.0f );
				Color c = color.get();
				float[] components = { c.getRed(), c.getGreen(), c.getBlue() };
				raster.setPixel( x, viewportDims.height - y - 1, components );

				// Feedback on how far along a render is.
				pixelsShaded++;

				currentPixel.add( widthStep );
			}

			pixel.add( heightStep );
			
			if( progressUpdater != null )
				progressUpdater.updateProgress( (int)( 100 * getProgress() ) );
		}
	}

	/**
	 * Casts a ray and calculates the color.
	 * 
	 * @param ray
	 * 			The ray to cast.
	 * @param ni
	 * 			The incident index of refraction.
	 * @param depth
	 * 			The current recursion depth. This starts at the maximum depth
	 * 			and is decremented each recursion.
	 * @param weight
	 * 			The weight of the specified ray.
	 * @return
	 * 			The color sampled by the ray.
	 */
	protected Color3f castRay( Ray3d ray, double ni, int depth, double weight )
	{

		// Early out if max recursion depth is reached.
		if( depth == 0 )
			return new Color3f();

		// Get the closest intersection
		Intersection intersections = new Intersection();
		getNearestObject( scene.getShapes(), ray, intersections );
		double time = intersections.time;
		Shape shape = intersections.shape;
		if( shape == null )
			return new Color3f();

		Point3d intersection = ray.interpolate( time );
		Vector3d normal = new Vector3d( shape.surfaceNormal( intersection ) );
		normal.normalize();

		// Negate normal if inside object.
		Vector3d incident = new Vector3d( ray.getDirection() );
		if( incident.dot( normal ) > 0.0 )
			normal.negate();

		// Calculate index of refraction and attenuation.
		Color3f attenuation = new Color3f();
		SurfaceProperties sp = shape.getSurfaceProperties();
		double nt = 0.0;
		
		// If incident index of refraction is 1, the ray is outside an object.
		if( ni == 1.0 )
		{
			nt = sp.getIndexOfRefraction();
			attenuation.set( scene.getAir().getAttenuation() );
		}
		// Otherwise it is inside an object.
		else
		{
			nt = 1.0;
			attenuation.set( sp.getAttenuation() );
		}

		// Get weights to determine how much of the ray is reflected and transmitted.
		double reflectionCoeff = calculateReflectionCoefficient( incident, normal, ni, nt );
		double transmissionCoeff = 1.0 - reflectionCoeff;
		reflectionCoeff = reflectionCoeff * sp.getSpecularCoefficient();
		transmissionCoeff = transmissionCoeff * sp.getSpecularCoefficient();

		// Apply local lighting
		Color3f pixelColor = new Color3f();
		applyLocalIllumination( shape, ray, reflectionCoeff, intersection, normal, pixelColor );

		/* Debugging */
		//pixelColor.set( object.getSurfaceProperties().diffuse );

		// Cast reflection rays recursively
		double reflectedWeight = reflectionCoeff * weight;
		if( reflectedWeight > scene.getMinimumRayWeight() )
		{
			Vector3d reflection = getReflection( incident, normal );
			Ray3d reflectionRay = new Ray3d( intersection, reflection );
			Color3f reflectionColor = castRay( reflectionRay, ni, depth - 1, reflectedWeight );
			reflectionColor.scale( (float) reflectionCoeff );
			pixelColor.add( reflectionColor );
		}

		// Cast transmission rays recursively
		double transWeight = transmissionCoeff * weight;
		if( transWeight > scene.getMinimumRayWeight() )
		{
			Vector3d transmission = getTransmission( incident, normal, ni, nt );
			Ray3d transmissionRay = new Ray3d( intersection, transmission );
			Color3f transmissionColor = castRay( transmissionRay, nt, depth - 1, transWeight );
			transmissionColor.scale( (float) transmissionCoeff );
			pixelColor.add( transmissionColor );
		}

		// Calculate the amount of attenuation.
		attenuation.set( (float)Math.pow( attenuation.x, time ),
				(float)Math.pow( attenuation.y, time ),
				(float)Math.pow( attenuation.z, time ) );

		// Attenuate the sampled color.
		pixelColor.set(
				attenuation.x * pixelColor.x,
				attenuation.y * pixelColor.y,
				attenuation.z * pixelColor.z
				);

		return pixelColor;
	}

	/**
	 * Returns the reflection coefficient.
	 * 
	 * @param incident
	 * 			The incident vector.
	 * @param normal
	 * 			The surface normal.
	 * @param ni
	 * 			The incident index of refraction.
	 * @param nt
	 * 			The transmission index of refraction.
	 * @return
	 * 			The reflection coefficient.
	 */
	protected static double calculateReflectionCoefficient( Vector3d incident, Vector3d normal, double ni, double nt )
	{	
		double iDotN = -incident.dot( normal );
		double refractionRatio = ni / nt;
		double radicand = 1.0 - ( refractionRatio * refractionRatio ) * ( 1.0 - iDotN * iDotN );
		if( radicand < 0.0 )
			return 1.0;

		double sqrtRadicand = Math.sqrt( radicand );
		double perpReflEnergy = refractionRatio * iDotN - sqrtRadicand;
		double perpIncidentEnergy = refractionRatio * iDotN + sqrtRadicand;
		double perpRatio = perpReflEnergy / perpIncidentEnergy;

		double paraReflEnergy = iDotN - refractionRatio * sqrtRadicand;
		double paraIncidentEnergy = iDotN + refractionRatio * sqrtRadicand;
		double paraRatio = paraReflEnergy / paraIncidentEnergy;

		double reflectionCoefficient = 0.5 * ( perpRatio * perpRatio + paraRatio * paraRatio );
		return reflectionCoefficient;
	}

	/**
	 * Applies local illumination to an object.
	 * 
	 * @param object
	 * 			The object to illuminate.
	 * @param ray
	 * 			The ray that intersecting the object.
	 * @param reflectionCoeff
	 * 			The reflection coefficient.
	 * @param intersection
	 * 			The intersection point of the ray with the object.
	 * @param normal
	 * 			The surface normal at the point of intersection.
	 * @param color
	 * 			The color of the object with illumination applied.
	 */
	protected void applyLocalIllumination( Shape object, Ray3d ray, double reflectionCoeff, Point3d intersection, Vector3d normal, Color3f color )
	{
		Color3f diffuse = new Color3f();
		Color3f specular = new Color3f();
		for( Light light : scene.getLights() )
		{
			Vector3d unitLightDir = new Vector3d();
			unitLightDir.sub( light.getPosition(), intersection );

			// single sample shadow ray cast
			float coverage = 1.0f;
			double distanceToLight = unitLightDir.length();
			unitLightDir.normalize();

			if( castShadowRay( scene, new Ray3d( intersection, unitLightDir ), distanceToLight ) )
				continue;

			double nDotL = unitLightDir.dot( normal );
			if( nDotL > 0.0 )
				diffuse.scaleAdd( coverage * (float)nDotL, light.getColor(), diffuse );

			Vector3d reflection = getReflection( unitLightDir, normal );
			Vector3d unitEyeDir = new Vector3d( ray.getDirection() );

			double rDotE = reflection.dot( unitEyeDir );
			if( rDotE > 0.0 )
			{
				double specularScalar = Math.pow( rDotE, object.getSurfaceProperties().getSpecularExponent() );
				specular.scaleAdd( (float) specularScalar, light.getColor(), specular );
			}
		}

		Color3f diffCoeffs = object.getSurfaceProperties().getDiffuse();
		diffuse.set( diffCoeffs.x * diffuse.x, diffCoeffs.y * diffuse.y, diffCoeffs.z * diffuse.z );
		color.add( diffuse );

		color.scaleAdd( (float) reflectionCoeff, specular, color );

		Color3f ambientLight = new Color3f();
		ambientLight.set( scene.getAmbientLight() );
		ambientLight.set( diffCoeffs.x * ambientLight.x, diffCoeffs.y * ambientLight.y, diffCoeffs.z * ambientLight.z );
		color.add( ambientLight );
	}

	/**
	 * Returns whether or not an object occludes a light.
	 * 
	 * @param scene
	 * 			The scene data.
	 * @param ray
	 * 			The ray to check for occluders along.
	 * @param distanceToLight
	 * 			The time at which the ray reaches the light.
	 * @return
	 * 			{@code true} if there is occlusion, otherwise {@code false}.
	 */
	protected boolean castShadowRay( Scene scene, Ray3d ray, double distanceToLight )
	{
		// TODO If the ray direction is the unnormalized vector between the light and the
		// intersection it is possible to compare against 1.0 since any object
		// intersection at a time less than 1.0 must be occluding the light.
		for( Shape object : scene.getShapes() )
		{
			double time = object.nearestIntersectionTime( ray );
			if( time != Intersection.NoIntersection )
			{
				if( time > Scene.EPSILON && time < distanceToLight + Scene.EPSILON  )
					return true;
			}
		}	
		return false;
	}

	/**
	 * Finds an {@code Intersection} for the closest object.
	 * 
	 * @param objects
	 * 			The objects to check for intersection.
	 * @param ray
	 * 			The ray to check for intersection.
	 * @param intersections
	 * 			The intersection of the nearest object.
	 */
	protected void getNearestObject( ArrayList<Shape> objects, Ray3d ray, Intersection intersections )
	{
		Shape nearestObject = null;
		double minTime = Double.POSITIVE_INFINITY;
		for( Shape object : objects )
		{
			double time = object.nearestIntersectionTime( ray );
			if( time != Intersection.NoIntersection && time < minTime && time > Scene.EPSILON )
			{
				nearestObject = object;
				minTime = time;
			}
		}

		intersections.time = minTime;
		intersections.shape = nearestObject;
	}

	/**
	 * Returns a reflection vector
	 * 
	 * @param incident
	 * 			The incident vector.
	 * @param normal
	 * 			The surface normal.
	 * @return
	 * 			The reflection vector.
	 */
	protected static Vector3d getReflection( Vector3d incident, Vector3d normal )
	{
		double iDotN = incident.dot( normal );
		Vector3d reflection = new Vector3d();
		reflection.scaleAdd( -2.0 * iDotN, normal, incident );
		return reflection;
	}

	/**
	 * Returns a transmission vector.
	 * 
	 * @param incident
	 * 			The incident vector.
	 * @param normal
	 * 			The surface normal.
	 * @param ni
	 * 			The incident index of refraction.
	 * @param nt
	 * 			The surface index of refraction.
	 * @return
	 * 			The transmission vector.
	 */
	protected static Vector3d getTransmission( Vector3d incident, Vector3d normal, double ni, double nt )
	{
		Vector3d transmission = new Vector3d();
		double iDotN = -incident.dot( normal );
		double refractionRatio = ni / nt;
		double radicand = 1.0 - ( refractionRatio * refractionRatio ) * ( 1.0 - iDotN * iDotN );
		if( radicand < 0.0 )
			return null;

		double sqrtRadicand = Math.sqrt( radicand );

		transmission.scale( refractionRatio * iDotN - sqrtRadicand, normal );
		transmission.scaleAdd( refractionRatio, incident, transmission );
		return transmission;
	}

	/**
	 * Get the image that is rendered to.
	 * 
	 * @return
	 * 			The image rendered to.
	 */
	public BufferedImage getBufferedImage()
	{
		return imageBuffer;
	}

	/**
	 * Returns the scene data.
	 * 
	 * @return
	 * 			The scene data.
	 */
	public Scene getScene()
	{
		return scene;
	}

	/**
	 * Returns the number of pixels shaded.
	 * 
	 * @return
	 * 			The number of pixels shaded.
	 */
	public int getPixelsShaded()
	{
		return pixelsShaded;
	}
	
	/**
	 * Returns the percentage of pixels shaded.
	 * 
	 * @return
	 * 			The percentage of pixels shaded.
	 */
	public float getProgress()
	{
		return (float)pixelsShaded / (float)(scene.getWidth() * scene.getHeight());
	}

	private BufferedImage imageBuffer;
	protected Scene scene;

	// The number of pixels shaded so far in the image
	private int pixelsShaded;
}
