import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.*;
import java.util.*;

import javax.imageio.*;
import javax.swing.*;
//import javax.swing.event.*;
import javax.swing.filechooser.FileFilter;

/**
 * The main GUI for the ray tracing application.
 */
public class RayTracerFrame extends JFrame
{
	/**
	 * Constructs a new application GUI.
	 * 
	 * @param defaultSceneFile
	 * 			The scene file that is loaded by default.
	 */
	public RayTracerFrame( String defaultSceneFile )
	{
		initComponents();
		
		scene = new Scene();
		renderTask = new RenderTask(null);
		renderTimer = new SystemTimer();
		
		// Load the specified scene file.
		if( null != defaultSceneFile )
		{
			currentFile = new File( defaultSceneFile ); 
			LoadFile( currentFile );
		}
	}

	private void initComponents()
	{
		setTitle( titleString );
		setSize( DEFAULT_FRAME_WIDTH, DEFAULT_FRAME_HEIGHT );
		
		// Center window.
		setLocationRelativeTo( null );
		
		// Set up the file chooser.
		fileChooser = new JFileChooser();
		fileChooser.setCurrentDirectory( new File( "./" ) );

		// Prevent the window from becoming too small.
		addComponentListener( new MinimumFrameSizeComponent( DEFAULT_FRAME_WIDTH, DEFAULT_FRAME_HEIGHT ) );

		// Create the File menu.
		JMenu fileMenu = new JMenu( "File" );
		fileMenu.setMnemonic( KeyEvent.VK_F );

		// Create the File > Open menu item.
		openAction = new OpenAction();
		
		// Create the File > Save Render... menu item.
		saveRenderAction = new SaveRenderAction();
		saveRenderAction.setEnabled( false );
		
		// Create the File > Reload menu item.
		JMenuItem reloadFileItem = new JMenuItem( "Reload", KeyEvent.VK_R);
		reloadFileItem.setToolTipText( "Reloads the current file from disk, discarding any changes." );
		reloadFileItem.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_F12, 0 ) );
		reloadFileItem.setIcon( new ImageIcon( "./resources/Refresh16.gif" ) );

		// Listener to reload the current file.
		reloadFileItem.addActionListener( new
				ActionListener()
				{
					public void actionPerformed( ActionEvent e )
					{
						LoadFile( currentFile );
					}
				}
		);

		// Create the File > Exit menu item.
		JMenuItem exitItem = new JMenuItem( "Exit", KeyEvent.VK_X );
		exitItem.addActionListener( new 
				ActionListener()
				{
					public void actionPerformed( ActionEvent e )
					{
						System.exit(0);
					}
				}
		);
		
		// Add menu items to File menu.
		fileMenu.add( openAction );
		fileMenu.add( saveRenderAction );
		fileMenu.add( reloadFileItem );
		fileMenu.add( exitItem );
		
		// Create the Render menu.
		JMenu renderMenu = new JMenu( "Render" );
		
		// Create the Render > Start Render menu item.
		startRenderAction = new StartRenderAction();
		
		// Create the Render > Stop Render menu item.
		stopRenderAction = new StopRenderAction();
		stopRenderAction.setEnabled( false );
		
		// Add menu items to Render menu.
		renderMenu.add( startRenderAction );
		renderMenu.add( stopRenderAction );

		// Create the Options menu.
		JMenu optionsMenu = new JMenu( "Options" );
		
		// Create the Options > Show Render Window menu item.
		toggleRenderWindowAction = new ToggleRenderWindowAction();
		JCheckBoxMenuItem toggleRenderWindowItem = new JCheckBoxMenuItem( toggleRenderWindowAction );
		toggleRenderWindowItem.setSelected( false );

		// Create the Options > Scene Properties menu item, and the dialog it displays.
		JMenuItem scenePropertiesItem = new JMenuItem( "Scene Properties..." );
		scenePropertiesItem.setIcon( new ImageIcon( "./resources/Properties16.gif" ) );
		scenePropertiesDialog = new ScenePropertiesDialog( this );
		scenePropertiesItem.addActionListener( new
				ActionListener()
				{
					public void actionPerformed( ActionEvent e )
					{
						// Displays the scene properties dialog and gets updated scene information.
						scenePropertiesDialog.setViewportWidth( scene.getWidth() );
						scenePropertiesDialog.setViewportHeight( scene.getHeight() );
						scenePropertiesDialog.setRecursionDepth( scene.getRecursionDepth() );
						scenePropertiesDialog.setMinimumRayWeight( scene.getMinimumRayWeight() );

						int result = scenePropertiesDialog.showDialog();
						if( result == JOptionPane.OK_OPTION )
						{
							// Retrieve data from dialog.
							scene.setViewportDimensions( scenePropertiesDialog.getViewportWidth(), scenePropertiesDialog.getViewportHeight(), 0 );
							scene.setRecursionDepth( scenePropertiesDialog.getRecursionDepth() );
							scene.setMinimumRayWeight( scenePropertiesDialog.getMinimumRayWeight() );
						}
					}
				}
		);
		
		// Add menu items to Options menu.
		optionsMenu.add( toggleRenderWindowItem );
		optionsMenu.add( scenePropertiesItem );
		
		// Create the Help menu.
		JMenu helpMenu = new JMenu( "Help" );
		helpMenu.setMnemonic( KeyEvent.VK_H );

		// Create the Help > Controls menu item.
		JMenuItem controlsItem = new JMenuItem( "Controls", KeyEvent.VK_C );

		// Create the Help > About menu item.
		JMenuItem aboutItem = new JMenuItem( "About", KeyEvent.VK_A );
		aboutItem.addActionListener( new
				ActionListener()
				{
					public void actionPerformed( ActionEvent e )
					{
						JOptionPane.showMessageDialog( null, aboutString );
					}
				}
		);

		// Add menu items to Help menu.
		helpMenu.add( controlsItem );
		helpMenu.add( aboutItem );
		
		// Create menu bar and add menus to it.
		JMenuBar menuBar = new JMenuBar();
		menuBar.add( fileMenu );
		menuBar.add( renderMenu );
		menuBar.add( optionsMenu );
		menuBar.add( helpMenu );
		setJMenuBar( menuBar  );
		
		// Create tool bar.
		JToolBar toolBar = new JToolBar();
		
		// Create tool bar Open button.		
		JButton openButton = new JButton( openAction );
		openButton.setHideActionText( true );
		
		// Create tool bar Save Render button.
		JButton saveRenderButton = new JButton( saveRenderAction );
		saveRenderButton.setHideActionText( true );
		
		// Create tool bar Start Render button.
		JButton startRenderButton = new JButton( startRenderAction );
		startRenderButton.setHideActionText( true );
		
		// Create tool bar Stop Render button.
		JButton stopRenderButton = new JButton( stopRenderAction );
		stopRenderButton.setHideActionText( true );
		
		// Create tool bar Toggle Render Window button.
		JToggleButton toggleRenderWindowButton = new JToggleButton( toggleRenderWindowAction );
		toggleRenderWindowButton.setHideActionText( true );
		
		// Add tool bar items.
		toolBar.setFloatable( false );
		toolBar.add( openButton );
		toolBar.add(  saveRenderButton );
		toolBar.addSeparator();
		toolBar.add( startRenderButton );
		toolBar.add( stopRenderButton );
		toolBar.addSeparator();
		toolBar.add(  toggleRenderWindowButton );
		add( toolBar, BorderLayout.PAGE_START );
		
		// Layout the content pane of the application.
		JPanel mainPanel = new JPanel();
		GridBagLayout gridBagLayout = new GridBagLayout();
		mainPanel.setLayout( gridBagLayout );
		getContentPane().add( mainPanel );

		// Add the text area.
		sceneTextArea = new JTextArea();
		GridBagConstraints objectPanelConstraints = new GridBagConstraints();
		objectPanelConstraints.gridx = 0;
		objectPanelConstraints.gridy = 0;
		objectPanelConstraints.gridwidth = 1;
		objectPanelConstraints.gridheight = 1;
		objectPanelConstraints.weightx = 100;
		objectPanelConstraints.weighty = 100;
		objectPanelConstraints.fill = GridBagConstraints.BOTH;
		objectPanelConstraints.anchor = GridBagConstraints.NORTH;
		mainPanel.add( new JScrollPane( sceneTextArea ), objectPanelConstraints );
//		sceneTextArea.getDocument().addDocumentListener( new
//				DocumentListener()
//				{
//					public void insertUpdate( DocumentEvent e ) { sceneChange = true; }
//					public void removeUpdate( DocumentEvent e ) { sceneChange = true; }
//					public void changedUpdate( DocumentEvent e ) {}
//				}
//		);

		// Add status bar with progress bar.
		JPanel statusPanel = new JPanel( new FlowLayout( FlowLayout.LEFT, 10, 2 ) );
		progressBar = new JProgressBar();
		progressBar.setStringPainted( true );
		statusText = new JLabel( " " );
		statusPanel.add( progressBar );	
		statusPanel.add( statusText );
		GridBagConstraints statusPanelConstraints = new GridBagConstraints();
		statusPanelConstraints.gridx = 0;
		statusPanelConstraints.gridy = 1;
		statusPanelConstraints.gridwidth = 1;
		statusPanelConstraints.gridheight = 1;
		statusPanelConstraints.weightx = 100;
		statusPanelConstraints.weighty = 0;
		statusPanelConstraints.fill = GridBagConstraints.HORIZONTAL;
		statusPanelConstraints.anchor = GridBagConstraints.WEST;
		mainPanel.add( statusPanel, statusPanelConstraints );
		
		// Create the frame to display the rendered image.
		renderFrame = new RenderFrame();
		renderFrame.addWindowListener( new
				WindowAdapter()
				{
					@Override
					public void windowActivated( WindowEvent e )
					{
						toggleRenderWindowAction.putValue( Action.SELECTED_KEY, Boolean.TRUE );
						super.windowActivated( e );
					}

					@Override
					public void windowClosing( WindowEvent e )
					{
						// Hide this window instead of closing it.
						renderFrame.setVisible( false );
						
						// Keep our GUI in sync with the window state.
						toggleRenderWindowAction.putValue( Action.SELECTED_KEY, Boolean.FALSE );
					}
				}
		);
		
		renderFrame.setVisible( false );
	}

	// Encapsulates opening scene files through a file chooser.
	private class OpenAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;

		public OpenAction()
		{
			putValue( Action.NAME, "Open" );
			putValue( Action.MNEMONIC_KEY, KeyEvent.VK_O );
			putValue( Action.SHORT_DESCRIPTION, "Open a scene file." );
			putValue( Action.SMALL_ICON, new ImageIcon( "./resources/Open16.gif" ) );
			putValue( Action.LARGE_ICON_KEY, new ImageIcon( "./resources/Open24.gif" ) );
			
		}

		public void actionPerformed( ActionEvent event )
		{
			// Filter out everything but directories and txt files.
			fileChooser.resetChoosableFileFilters();
			fileChooser.setFileFilter( new
					FileFilter()
					{
						public boolean accept( File file )
						{
							boolean isTxt = file.getName().toLowerCase().endsWith(".txt");
							return isTxt || file.isDirectory();
						}
		
						public String getDescription() { return "(*.txt) Scene Files"; }
					}
			);

			// Show the file dialog and load scene if possible.
			int openResult = fileChooser.showOpenDialog( null );
			if( openResult == JFileChooser.APPROVE_OPTION )
			{
				currentFile = fileChooser.getSelectedFile();
				LoadFile( currentFile );
			}
			else if( openResult == JFileChooser.ERROR_OPTION )
			{
				statusText.setText( "Error opening file." );
			}
		}
	}
	
	// Encapsulates saving a rendered image.
	private class SaveRenderAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;
		
		public SaveRenderAction()
		{
			putValue( Action.NAME, "Save Render...");
			putValue( Action.MNEMONIC_KEY, KeyEvent.VK_T );
			putValue( Action.SHORT_DESCRIPTION, "Save a rendered image." );
			putValue( Action.SMALL_ICON, new ImageIcon( "./resources/Save16.gif" ) );
			putValue( Action.LARGE_ICON_KEY, new ImageIcon( "./resources/Save24.gif" ) );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			assert fileChooser != null;
			assert statusText != null;
			
			// Make sure we have a render that can be saved.
			RenderedImage image = (RenderedImage)renderFrame.getImage();

			assert image != null;
			
			// Filter out everything but directories and bmp files.
			fileChooser.resetChoosableFileFilters();
			fileChooser.setFileFilter( new
					FileFilter()
					{
						public boolean accept( File file )
						{
							return file.getName().toLowerCase().endsWith(".bmp") || file.isDirectory();
						}

						public String getDescription() { return "(*.bmp) Bitmap Files"; }
					}
			);
			
			// Suggest a save file for the user.
			fileChooser.setSelectedFile( new File( getSuggestedSaveFileName( currentFile.getName() ) ) );

			// Write bmp image if possible.
			int openResult = fileChooser.showSaveDialog( null );
			if( openResult == JFileChooser.APPROVE_OPTION )
			{
				File saveFile = fileChooser.getSelectedFile();
				if( !saveFile.getName().endsWith( ".bmp" ) )
					saveFile = new File( saveFile.getName() + ".bmp ");
				
				try
				{
					saveRender( image, saveFile );
					statusText.setText( "Image saved succesfully: " + saveFile.getAbsolutePath() );
				}
				catch( IOException ex )
				{
					statusText.setText( "Error: Image could not be saved.\n" + ex.getMessage() );
				}
			}
			else if( openResult == JFileChooser.ERROR_OPTION )
			{
				statusText.setText( "Error loading file." );
			}
		}
		
		public String getSuggestedSaveFileName( String inFileName )
		{
			assert inFileName != null;
			
			String suggestedName;
			String extension = ".bmp";
			int extensionIndex = inFileName.lastIndexOf( '.' );
			if( -1 != extensionIndex )
			{
				suggestedName = inFileName.substring( 0, extensionIndex ) + extension;
			}
			else
			{
				suggestedName = inFileName + extension;
			}
			
			return suggestedName;			
		}
	}

	// Encapsulates starting a render.
	private class StartRenderAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;

		public StartRenderAction()
		{
			putValue( Action.NAME, "Start Render" );
			putValue( Action.MNEMONIC_KEY, KeyEvent.VK_R );
			putValue( Action.SHORT_DESCRIPTION, "Start rendering the scene." );
			putValue( Action.SMALL_ICON, new ImageIcon( "./resources/Start16.gif" ) );
			putValue( Action.LARGE_ICON_KEY, new ImageIcon( "./resources/Start24.gif" ) );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			assert renderTask != null;
			
			if( !renderTask.isStarted() )
			{
				startRenderTask();
			}
		}
	}
	
	// Encapsulates stopping a render.
	private class StopRenderAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;

		public StopRenderAction()
		{
			putValue( Action.NAME, "Stop Render" );
			putValue( Action.MNEMONIC_KEY, KeyEvent.VK_P );
			putValue( Action.SHORT_DESCRIPTION, "Stop rendering the scene." );
			putValue( Action.SMALL_ICON, new ImageIcon( "./resources/Stop16.gif" ) );
			putValue( Action.LARGE_ICON_KEY, new ImageIcon( "./resources/Stop24.gif" ) );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			assert renderTask != null;
			
			if( renderTask.isStarted() )
			{
				cancelRenderTask();
			}
		}
	}
	
	// Encapsulates toggling the window showing the rendered image.
	private class ToggleRenderWindowAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;
		
		public ToggleRenderWindowAction()
		{
			putValue( Action.NAME, showString);
			putValue( Action.MNEMONIC_KEY, KeyEvent.VK_W );
			putValue( Action.SHORT_DESCRIPTION, showString );
			putValue( Action.SMALL_ICON, new ImageIcon( "./resources/RenderWindow16.gif" ) );
			putValue( Action.LARGE_ICON_KEY, new ImageIcon( "./resources/RenderWindow24.gif" ) );
			putValue( Action.SELECTED_KEY, Boolean.FALSE );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			assert renderFrame != null;
			
			boolean isSelected = (Boolean)getValue( Action.SELECTED_KEY );
			renderFrame.setExtendedState( Frame.NORMAL );
			renderFrame.setVisible( isSelected );

			String toolTip = isSelected ? hideString : showString;
			putValue( Action.SHORT_DESCRIPTION, toolTip );
		}
		
		private String showString = "Show Render Window";
		private String hideString = "Hide Render Window";
	}
	
	/**
	 * Loads a scene file from disk.
	 * 
	 * @param file
	 * 			The scene file to load.
	 */
	public void LoadFile( File file )
	{
		assert file != null;
		assert scene != null;
		assert renderFrame != null;
		
		try
		{
			setTextFromFile( sceneTextArea, file );
			boolean success = scene.LoadScene( sceneTextArea.getText() );

			if( success )
				statusText.setText( "Scene loaded successfully: " + file.getAbsolutePath() );
			else
				statusText.setText( "File may not have loaded properly (I.e. bad data or unknown types).");
		}
		catch( FileNotFoundException e )
		{
			String message = "File could not be found: " + file.getAbsolutePath();
			JOptionPane.showMessageDialog( null, message );
			statusText.setText( message );
		}
	}

	
	/**
	 * Populates a text area from a file.
	 * 
	 * @param textArea
	 * 			The text area to populate.
	 * @param file
	 * 			The file to load text from.
	 * 
	 * @throws FileNotFoundException
	 */
	public static void setTextFromFile( JTextArea textArea, File file ) throws FileNotFoundException
	{
		assert textArea != null;
		assert file != null;
		
		textArea.replaceRange( "", 0, textArea.getText().length() );
		Scanner in = new Scanner( new FileInputStream( file ) );
		while( in.hasNextLine() )
		{
			textArea.append( in.nextLine() );
			textArea.append( "\n" );
		}
	}

	/**
	 * Saves an image to a bitmap file.
	 * 
	 * @param image
	 * 			The image to be saved.
	 * @param saveFile
	 * 			The file to save the image to.
	 * 
	 * @throws IOException
	 */
	private void saveRender( RenderedImage image, File saveFile ) throws IOException
	{
		// TODO Check to make sure the image was successfully saved rather than assuming it.
		ImageIO.write( image, "bmp", saveFile );
	}
	
	/**
	 * Centers a component on the screen.
	 * 
	 * @param comp
	 * 			The component to center.
	 */
	public static void centerComponentOnScreen( Component comp )
	{
		Toolkit toolkit = Toolkit.getDefaultToolkit();
		Dimension screenSize = toolkit.getScreenSize();
		int frameX = (screenSize.width / 2) - (comp.getWidth() / 2);
		int frameY = (screenSize.height / 2) - (comp.getHeight() / 2);
		comp.setBounds( frameX, frameY, comp.getWidth(), comp.getHeight() );
	}
	
	// TODO Add a way for the user to change super-sampling options.
	private void startRenderTask()
	{
		assert scene != null;
		
		// Clone the Scene so we can edit in the GUI while rendering.
		Scene sceneCopy = (Scene)scene.clone();
		RayTracerUniformSuperSampler rayTracer = new RayTracerUniformSuperSampler( sceneCopy, 1 );
		renderFrame.setImage( rayTracer.getBufferedImage() );
		renderTask = new RenderTask( rayTracer );
		
		// Add listener to recieve progress updates.
		renderTask.addPropertyChangeListener( new
				PropertyChangeListener()
				{
					public void propertyChange(PropertyChangeEvent evt)
					{
						if ("progress".equals(evt.getPropertyName()))
						{
							progressBar.setValue((Integer)evt.getNewValue());
							renderFrame.repaint();
							statusText.setText( "Elapsed render time: " + renderTimer.toString() + " (HH:MM:SS)" );
						}
					}
				}
		);
		
		saveRenderAction.setEnabled( false );
		startRenderAction.setEnabled( false );
		stopRenderAction.setEnabled( true );
		centerComponentOnScreen( renderFrame );
		renderFrame.setVisible( true );
		renderTimer.start();
		
		renderTask.execute();
	}

	private void endRenderTask()
	{
		renderFrame.repaint();
		progressBar.setValue( 0 );
		statusText.setText( "Elapsed render time: " + renderTimer.toString() + " (HH:MM:SS)" );
		saveRenderAction.setEnabled( true );
		stopRenderAction.setEnabled( false );
		startRenderAction.setEnabled( true );
	}
	
	private void cancelRenderTask()
	{
		if( !renderTask.isDone() )
			renderTask.cancel( true );
		
		assert renderTask.isDone();
	}

	/**
	 * A background task to handle rendering.
	 */
	public class RenderTask extends SwingWorker< Object, Object > implements IProgressable
	{
		public RenderTask( RayTracer rayTracer )
		{
			this.rayTracer = rayTracer;
		}

		@Override
		public Object doInBackground() throws Exception
		{
			rayTracer.render( this );
			return null;
		}

		public void updateProgress( int progress )
		{
			assert 0 <= progress && progress <= 100;
			setProgress( progress );
		}

		@Override
		protected void done()
		{
			endRenderTask();
			super.done();
		}
		
		public boolean isStarted()
		{
			return SwingWorker.StateValue.STARTED == getState();
		}
		
		private RayTracer rayTracer;
	}
	
	// Fields
	private RenderFrame renderFrame;
	private ScenePropertiesDialog scenePropertiesDialog;
	
	private JTextArea sceneTextArea;
	private JLabel statusText;
	private JProgressBar progressBar;
	private JFileChooser fileChooser;
	
	private OpenAction openAction;
	private SaveRenderAction saveRenderAction;
	private StartRenderAction startRenderAction;
	private StopRenderAction stopRenderAction;
	private ToggleRenderWindowAction toggleRenderWindowAction;
	
	private RenderTask renderTask;
	private SystemTimer renderTimer;
	
	private File currentFile;
	private Scene scene;

	private static final int DEFAULT_FRAME_WIDTH = 640;
	private static final int DEFAULT_FRAME_HEIGHT = 480;

	private static final String aboutString = "CS400 Raytracer\nDigipen Institute of Technology (c) 2007\nKevin Tomlinson";
	private static final String titleString = "RayTracer";
	
	private static final long serialVersionUID = 1L;
}
