import javax.vecmath.*;

/**
 * Implements a ray tracer that provides uniform super-sampling.
 * 
 * The grid size determines how many rays are cast per pixel. A value of 1
 * will cast one ray at the center of the pixel. A value of 2 will split the
 * pixel into a 2x2 grid and cast four rays. A value of 3 splits the pixel into
 * a 3x3 grid, and so on.
 */
public class RayTracerUniformSuperSampler extends RayTracer
{
	/**
	 * Constructs a uniform super-sampling ray tracer.
	 * 
	 * @param scene
	 * 			The scene to render.
	 * @param gridSize
	 * 			The size of the grid to super-sample with.
	 */
	public RayTracerUniformSuperSampler( Scene scene, int gridSize )
	{
		super( scene );
		this.gridSize = gridSize;
	}

	public void calculateSamplingInfo( Point3d pixel )
	{
		Camera camera = scene.getCamera();

		pixel.set( camera.getViewCenter() );
		pixel.sub( camera.getWidthVector() );
		pixel.sub( camera.getHeightVector() );
	}

	public Color3f samplePixel( Point3d startPixel, Vector3d widthStep, Vector3d heightStep )
	{
		// Calculate step sizes
		Vector3d subPixelWidth = new Vector3d();
		Vector3d subPixelHeight = new Vector3d();
		subPixelWidth.scale( 1.0 / gridSize, widthStep );
		subPixelHeight.scale( 1.0 / gridSize, heightStep );

		// Calculate first sub-pixel position.
		Point3d subPixel = new Point3d( startPixel );
		subPixel.scaleAdd( 0.5, subPixelWidth, subPixel );
		subPixel.scaleAdd( 0.5, subPixelHeight, subPixel );

		Point3d eyePoint = new Point3d();
		eyePoint.add( scene.getCamera().getViewCenter(), scene.getCamera().getEyeVector() );

		// Cast rays for each sub-pixel and accumulate the color.
		Vector3d rayDirection = new Vector3d();
		Color3f color = new Color3f();
		Ray3d ray = new Ray3d();
		Vector3d currentSubPixel = new Vector3d();
		for( int y = 0; y < gridSize; ++y )
		{
			currentSubPixel.set( subPixel );
			for( int x = 0; x < gridSize; ++x )
			{
				rayDirection.sub( currentSubPixel, eyePoint );
				ray.setNormalized( eyePoint, rayDirection );
				color.add( castRay( ray, 1.0, scene.getRecursionDepth(), 1.0 ) );

				currentSubPixel.add( subPixelWidth );
			}

			subPixel.add( subPixelHeight );
		}

		// Normalize color.
		color.scale( 1.0f / ( gridSize * gridSize ) );	
		return color;
	}

	private int gridSize;
}