import java.awt.Dimension;
import java.io.*;
import java.util.*;

import javax.vecmath.Color3f;

/**
 * A class that contains all information necessary to render a scene, including
 * objects in the scene, lights in the scene, viewport dimensions, and global 
 * scene properties.
 */
public class Scene implements CloneableEx
{
	/**
	 * Constructs the default (empty) scene.
	 */
	public Scene()
	{
		shapeList = new ArrayList<Shape>();
		lightList = new ArrayList<Light>();
		camera = new Camera();
		viewportDimensions = new Dimension( 1, 1 );
		ambientLight = new Color3f();
		recursionDepth = DEFAULT_RECURSION_DEPTH;
		minimumRayWeight = DEFAULT_MIN_RAY_WEIGHT;
		air = new Air();
	}

	protected Scene( Scene scene )
	{
		shapeList = Utilities.deepCopy( scene.getShapes() );
		lightList = Utilities.deepCopy( scene.getLights() );
		viewportDimensions = (Dimension)scene.getViewportDimensions().clone();
		camera = (Camera)scene.getCamera().clone();
		ambientLight = (Color3f)scene.getAmbientLight().clone();
		recursionDepth = scene.getRecursionDepth();
		minimumRayWeight = scene.getMinimumRayWeight();
		air = (Air)scene.getAir().clone();
	}
	
	@Override
	public Object clone()
	{
		return new Scene( this );
	}
	
	/**
	 * Resets the scene back to the default (empty) state.
	 */
	public void reset()
	{
		shapeList.clear();
		lightList.clear();
		camera = new Camera();
		viewportDimensions.setSize( 1, 1 );
		ambientLight = new Color3f();
		recursionDepth = DEFAULT_RECURSION_DEPTH;
		minimumRayWeight = DEFAULT_MIN_RAY_WEIGHT;
		air = new Air();		
	}
	
	/**
	 * Loads a scene from the specified file.
	 * 
	 * All previous scene data is discarded.
	 * 
	 * @param file
	 * 			The file containing scene data.
	 * @return
	 * 			{@code true} if the scene was loaded properly, otherwise {@code false}.
	 * 
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public boolean LoadScene( File file ) throws FileNotFoundException, IOException
	{
		reset();

		boolean success = false;
		DataParser dataParser = new DataParser();
		success = dataParser.parseFile( file.getAbsolutePath(), DELIMITER );
		if( success )
			success = LoadObjects( dataParser );

		return success;
	}

	/**
	 * Loads a scene from a string containing the scene data.
	 * 
	 * All previous scene data is discarded.
	 * 
	 * @param data
	 * 			The string containing the scene data.
	 * @return
	 * 			{@code true} if the scene was loaded properly, otherwise {@code false}.
	 */
	public boolean LoadScene( String data )
	{
		reset();

		boolean success = false;
		DataParser dataParser = new DataParser();
		success = dataParser.parseString( data, DELIMITER );
		if( success )
			success = LoadObjects( dataParser );

		return success;	
	}

	/**
	 * Loads scene data from the {@code DataParser}.
	 * 
	 * @param dataParser
	 * 			The parsed data.
	 * @return
	 * 			{@code true} if the scene was loaded properly, otherwise {@code false}.
	 */
	private boolean LoadObjects( DataParser dataParser )
	{
		boolean dataLoadedProperly = true;

		try
		{
			while( dataParser.tokensRemaining() )
			{
				String token = dataParser.getString();
				if( token.equals( "" ) )
				{
					// Sometimes a blank line is tokenized. So skip it.
					continue;
				}
				else if( token.equals( "SPHERE" ) )
				{
					Shape shape = new Sphere();
					shape.load( dataParser );
					shapeList.add( shape );
				}
				else if( token.equals( "BOX" ) )
				{
					Shape shape = new Box();
					shape.load( dataParser );
					shapeList.add( shape );
				}
				else if( token.equals( "POLYGON" ) )
				{
					Shape shape = new Polygon();
					shape.load( dataParser );
					shapeList.add( shape );
				}
				else if( token.equals( "ELLIPSOID" ) )
				{
					Shape shape = new Ellipsoid();
					shape.load( dataParser );
					shapeList.add( shape );
				}
				else if( token.equals( "LIGHT" ) )
				{
					Light light = new Light();
					light.load( dataParser );
					lightList.add( light );
				}
				else if( token.equals( "CAMERA" ) )
				{
					camera = new Camera();
					camera.load( dataParser );

					float viewportWidth = (float)camera.getWidthVector().length();
					float viewportHeight = (float)camera.getHeightVector().length();
					float aspectRatio =  viewportWidth / viewportHeight;
					setViewportDimensions( DEFAULT_RENDER_WIDTH, 0, aspectRatio );
				}
				else if( token.equals( "AMBIENT" ) )
				{
					ambientLight.set( dataParser.getFloat(), dataParser.getFloat(), dataParser.getFloat() );
				}
				else if( token.equals( "AIR" ) )
				{
					air.load( dataParser );
				}
				else
				{
					dataLoadedProperly = false;
				}

			}
		}
		catch( NumberFormatException e )
		{
			// Either some of the tokenized data is bad, the tokenized data
			// is incomplete, or an object tried to load itself incorrectly.
			dataLoadedProperly = false;
		}

		return dataLoadedProperly;
	}

	/**
	 * Returns the shapes in the scene.
	 * 
	 * @return
	 * 			The shapes in the scene.
	 */
	public ArrayList<Shape> getShapes() { return shapeList; }
	
	/**
	 * Returns the lights in the scene.
	 * 
	 * @return
	 * 			The lights in the scene.
	 */
	public ArrayList<Light> getLights() { return lightList; }
	
	/**
	 * Returns the ambient light in the scene.
	 * 
	 * @return
	 * 			The ambient light in the scene.
	 */
	public Color3f getAmbientLight() { return ambientLight; }

	/**
	 * Returns the scene's camera.
	 * 
	 * @return
	 * 			The scene's camera.
	 */
	public Camera getCamera() { return camera; }

	/**
	 * Returns the dimensions of the viewport in pixels.
	 * 
	 * @return
	 * 			The dimensions of the viewport.
	 */
	public Dimension getViewportDimensions() { return viewportDimensions; }
	
	/**
	 * Sets the dimensions of the viewport.
	 * 
	 * If {@code aspectRatio} is non-zero the dimension with the greatest value
	 * is used to determine the other dimensions aspect correct value.
	 * 
	 * Otherwise, if both dimensions are non-zero the viewport dimensions are
	 * set to those values.
	 * 
	 * Using negative values for any parameter is undefined.
	 * 
	 * @param width
	 * 			The width of the viewport in pixels.
	 * 
	 * @param height
	 * 			The height of the viewport in pixels.
	 * 
	 * @param aspectRatio
	 * 			The aspect ratio of the viewport.
	 */
	public void setViewportDimensions( int width, int height, float aspectRatio )
	{
		if( aspectRatio != 0 )
		{
			if( width > height )
			{
				viewportDimensions.setSize( width, (int)(width / aspectRatio) );
			} 
			else
			{
				viewportDimensions.setSize( (int)(height * aspectRatio), height );
			}
		}
		else
		{
			if( width != 0 && height != 0 )
				viewportDimensions.setSize( width, height );
		} 
	}

	/**
	 * Returns the width of the viewport in pixels.
	 * 
	 * @return
	 * 			The width of the viewport.
	 */
	public int getWidth() { return viewportDimensions.width; }
	
	/**
	 * Returns the height of the viewport in pixels.
	 * 
	 * @return
	 * 			The height of the viewport.
	 */
	public int getHeight() { return viewportDimensions.height; }

	/**
	 * Sets the maximum recursion depth of a traced ray.
	 * 
	 * @param depth
	 * 			The number of recursions a ray will traverse.
	 */
	public void setRecursionDepth( int depth ) { recursionDepth = depth; }
	
	/**
	 * Returns the maximum recursion depth of a traced ray.
	 * 
	 * @return
	 * 			The maximum recursion depth.
	 */
	public int getRecursionDepth() { return recursionDepth; }

	/**
	 * Sets the minimum weight of a ray.
	 * 
	 * If the weight of a ray falls below this minimum it will be discarded even
	 * if the maximum recursion depth has not yet been reached.
	 * 
	 * @param minWeight
	 * 			The minimum ray weight.
	 */
	public void setMinimumRayWeight( float minWeight ) { minimumRayWeight = minWeight; }
	
	/**
	 * Returns the minimum weight of a ray.
	 * 
	 * @return
	 * 			The minimum ray weight.
	 */
	public float getMinimumRayWeight() { return minimumRayWeight; }

	/**
	 * Sets the medium rays travel through when not inside an object.
	 * 
	 * @param air
	 * 			The air to set. 
	 */
	public void setAir( Air air ) { this.air = air; }
	
	/**
	 * Returns the medium rays travel through when not in an object.
	 * 
	 * @return
	 * 			The scene's air.
	 */
	public Air getAir() { return air; }

	private ArrayList<Shape> shapeList;
	private ArrayList<Light> lightList;
	private Color3f ambientLight;
	private Air air;
	private Camera camera;
	private Dimension viewportDimensions;
	private int recursionDepth;
	private float minimumRayWeight;

	public static final int DEFAULT_RECURSION_DEPTH = 10;
	public static final double EPSILON = 1e-6;
	public static final float DEFAULT_MIN_RAY_WEIGHT = 1e-2f;

	private static final int DEFAULT_RENDER_WIDTH = 400;
	@SuppressWarnings("unused")
	private static final int DEFAULT_RENDER_HEIGHT = 400;

	// Regular expression string to tokenize the input file.
	// Skips ')' , '(' , ',' , and ") (", skipping any whitespace before, after
	// or in between the character(s).
	private static final String DELIMITER = "\\s*\\)\\s*\\(\\s*|\\s*\\(\\s*|\\s*\\)\\s*|\\s*\\,\\s*|\\s+";
}
