import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A dialog to allow the user to change scene properties.
 * 
 * A user should set the properties of the dialog before calling {@code showDialog}
 * to populate the fields of the dialog.
 * 
 * If {@code showDialog} returns {@code JOptionPane.OK_OPTION} then the new
 * properties can be retrieved through the {@code get} methods.
 */
public class ScenePropertiesDialog extends JDialog
{
	public ScenePropertiesDialog( Frame parent )
	{	
		super( parent, true );
		
		initializeComponents();
		setLocationRelativeTo( parent );
	}

	private void initializeComponents()
	{
		setLayout( new BorderLayout() );

		// Create input boxes and labels.
		JPanel panel = new JPanel();
		panel.setLayout( new GridLayout( 0, 2, 5, 5 ) );
		panel.add( new JLabel( "Horizontal Resolution: " ) );
		panel.add( viewportWidthText );
		panel.add( new JLabel( "Vertical Resolution: " ) );
		panel.add( viewportHeightText );
		panel.add( new JLabel( "Recursion Depth: " ) );
		panel.add( recursionDepthText );
		panel.add( new JLabel( "Minimum Ray Weight: " ) );
		panel.add( minimumRayWeightText );	
		add( panel, BorderLayout.CENTER );

		// Create buttons.
		JButton okButton = new JButton( "Ok" );
		okButton.addActionListener( new
				ActionListener()
				{
					public void actionPerformed( ActionEvent e )
					{
						try
						{
							// Validate that all values are valid numbers.
							viewportWidth = Integer.parseInt( viewportWidthText.getText() );
							viewportHeight = Integer.parseInt( viewportHeightText.getText() );
							recursionDepth = Integer.parseInt( recursionDepthText.getText() );
							minimumRayWeight = Float.parseFloat( minimumRayWeightText.getText() );
		
							returnValue = JOptionPane.OK_OPTION;
							setVisible( false );
						}
						catch( NumberFormatException ex )
						{
							setTitle( "Invalid " + ex.getMessage() );
						}
					}
				}
		);

		JButton cancelButton = new JButton( "Cancel" );
		cancelButton.addActionListener( new
				ActionListener()
				{
					public void actionPerformed( ActionEvent e )
					{
						returnValue = JOptionPane.CANCEL_OPTION;
						setVisible( false );
					}
				}
		);

		JPanel buttonPanel = new JPanel();
		buttonPanel.add( okButton );
		buttonPanel.add( cancelButton );
		add( buttonPanel, BorderLayout.SOUTH );
		
		getRootPane().setDefaultButton( okButton );
		
		pack();
	}
	
	/**
	 * Shows the scene properties dialog.
	 * 
	 * @return
	 * 			{@code JOptionPane.OK_OPTION} if the user selected Ok, or
	 * 			{@code JOptionPane.CANCEL_OPTION} if the user canceled the dialog.
	 */
	public int showDialog()
	{
		setTitle("");
		viewportWidthText.setText( String.valueOf( getViewportWidth() ) );
		viewportHeightText.setText( String.valueOf( getViewportHeight() ) );
		recursionDepthText.setText( String.valueOf( getRecursionDepth() ) );
		minimumRayWeightText.setText( String.valueOf( getMinimumRayWeight() ) );
		returnValue = JOptionPane.CANCEL_OPTION;
		
		setVisible( true );
		
		return returnValue;
	}

	/**
	 * Populates the viewport width text box.
	 * 
	 * @param width
	 * 			The viewport width to be displayed.
	 */
	public void setViewportWidth( int width )
	{
		viewportWidth = width;
	}
	
	/**
	 * Returns the viewport width entered in the dialog.
	 * 
	 * @return
	 * 			The viewport width in pixels.
	 */
	public int getViewportWidth()
	{
		return viewportWidth;
	}
	
	/**
	 * Populates the viewport height text box.
	 * 
	 * @param height
	 * 			The viewport height to be displayed.
	 */
	public void setViewportHeight( int height )
	{
		viewportHeight = height;
	}
	
	/**
	 * Returns the viewport height entered in the dialog.
	 * 
	 * @return
	 * 			The viewport height in pixels.
	 */
	public int getViewportHeight()
	{
		return viewportHeight;
	}
	
	/**
	 * Populates the recursion depth text box.
	 * 
	 * @param depth
	 * 			The recursion depth to be displayed.
	 */
	public void setRecursionDepth( int depth )
	{
		recursionDepth = depth;
	}
	
	/**
	 * Returns the recursion depth entered in the dialog.
	 * 
	 * @return
	 * 			The recursion depth.
	 */
	public int getRecursionDepth()
	{
		return recursionDepth;
	}

	/**
	 * Populates the minimum ray weight text box.
	 * 
	 * @param weight
	 * 			The minimum ray weight to be displayed.
	 */
	public void setMinimumRayWeight( float weight )
	{
		minimumRayWeight = weight;
	}
	
	/**
	 * Returns the minimum ray weight entered in the dialog.
	 * 
	 * @return
	 * 			The minimum ray weight.
	 */
	public float getMinimumRayWeight()
	{
		return minimumRayWeight;
	}

	private JTextField viewportWidthText = new JTextField();
	private JTextField viewportHeightText = new JTextField();
	private JTextField recursionDepthText = new JTextField();
	private JTextField minimumRayWeightText = new JTextField();

	private int viewportWidth;
	private int viewportHeight;
	private int recursionDepth;
	private float minimumRayWeight;
	
	private int returnValue;
	
	private static final long serialVersionUID = 1L;
}
