import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * An abstract class that provides an interface for manipulating different
 * shapes.
 * <p>
 * The {@code Shape} class also provides a way to load and access the
 * {@link SurfaceProperties} of a shape.
 */
public abstract class Shape implements Persistent, CloneableEx
{
	/**
	 * Constructs a {@code Shape} with default surface properties.
	 */
	public Shape()
	{
		surfaceProperties = new SurfaceProperties();
	}

	/**
	 * Performs a deep copy of the specified {@code Shape} object.
	 * 
	 * @param shape
	 * 			The {@code Shape} object to copy.
	 */
	protected Shape( Shape shape )
	{
		surfaceProperties = (SurfaceProperties)shape.surfaceProperties.clone();
	}
	
	/**
	 * Constructs a {@code Shape} with the specified SurfaceProperties.
	 * 
	 * @param surfaceProperties
	 * 			The surface properties of this shape.
	 */
	public Shape( SurfaceProperties surfaceProperties )
	{
		this.surfaceProperties = surfaceProperties;
	}

	public abstract Object clone();
	
	/**
	 * Sets the surface properties.
	 * 
	 * @param surfaceProperties
	 * 			The surface properties to set.
	 */
	public void setSurfaceProperties( SurfaceProperties surfaceProperties )
	{
		this.surfaceProperties = surfaceProperties;
	}

	/**
	 * Returns the surface properties of this shape.
	 * 
	 * @return
	 * 			The surface properties.
	 */
	public SurfaceProperties getSurfaceProperties()
	{
		return surfaceProperties;
	}

	/**
	 * Gets the nearest intersection time between the specified ray and the
	 * shape.
	 * 
	 * @param ray
	 * 			The ray to intersect with the shape.
	 * @return
	 * 			The nearest time of intersection.
	 */
	public abstract double nearestIntersectionTime( Ray3d ray );

	/**
	 * Gets the surface normal at the specified point on the shape.
	 * 
	 * @param point
	 * 			The point on the surface of the shape.
	 * @return
	 * 			The surface normal of the point on the shape.
	 */
	public abstract Vector3d surfaceNormal( Point3d point );

	public void load( DataParser dataParser )
	{
		surfaceProperties.setDiffuse( dataParser.getColor3f() );	
		surfaceProperties.setSpecularCoefficient( dataParser.getDouble() );
		surfaceProperties.setSpecularExponent( dataParser.getDouble() );
		surfaceProperties.setAttenuation( dataParser.getColor3f() );
		surfaceProperties.setElectricPermittivity( dataParser.getDouble() );
		surfaceProperties.setMagneticPermeability( dataParser.getDouble() );
	}

	protected SurfaceProperties surfaceProperties;
}