import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * A class representing a sphere.
 * 
 * A sphere is specified by an origin and a radius.
 */
public class Sphere extends Shape
{
	/**
	 * Constructs the unit sphere at the origin.
	 */
	public Sphere()
	{
		setOrigin( new Point3d() );
		setRadius( 1.0 );
	}

	/**
	 * Performs a deep copy of the specified {@code Sphere} object.
	 * 
	 * @param sphere
	 * 			The {@code Sphere} object to copy.
	 */
	protected Sphere( Sphere sphere )
	{
		super( sphere );
		setOrigin( (Point3d)sphere.getOrigin().clone() );
		setRadius( sphere.getRadius() );
	}
	
	/**
	 * Constructs a sphere with the specified parameters.
	 * 
	 * @param origin
	 * 			The origin of the sphere.
	 * @param radius
	 * 			The radius of the sphere.
	 * @param sp
	 * 			The surface properties of the sphere.
	 */
	public Sphere( Vector3d origin, double radius, SurfaceProperties sp )
	{
		super( sp );
		this.setOrigin( new Point3d( origin ) );
		this.setRadius( radius );
	}

	@Override
	public Object clone()
	{
		return new Sphere( this );
	}

	public double nearestIntersectionTime( Ray3d ray )
	{
		double A = ray.getDirection().lengthSquared();
		Vector3d rayToSphere = new Vector3d();
		rayToSphere.sub( ray.getOrigin(), getOrigin() );
		double B = 2 * rayToSphere.dot( ray.getDirection() );
		double C = rayToSphere.lengthSquared() - ( getRadius() * getRadius() );	

		double discriminant = ( B * B ) - ( 4 * A * C );
		if( discriminant < 0.0 )
			return Intersection.NoIntersection;
		else
		{
			double sqrtdiscriminant = Math.sqrt( discriminant );
			double timeExit = ( -B + sqrtdiscriminant ) / ( 2 * A );
			double timeEnter = ( -B - sqrtdiscriminant ) / ( 2 * A );
			if( timeExit < Scene.EPSILON )
				return Intersection.NoIntersection;
			else if( timeEnter < Scene.EPSILON )
				return timeExit;
			else
				return timeEnter;
		}
	}

	public Vector3d surfaceNormal( Point3d point )
	{
		Vector3d normal = new Vector3d();
		normal.sub( point, this.getOrigin() );
		return normal;
	}

	public void load( DataParser dataParser )
	{
		setOrigin( dataParser.getPoint3d() );
		setRadius( dataParser.getDouble() );
		super.load( dataParser );
	}

	/**
	 * Sets the sphere's origin.
	 * @param origin The origin to set.
	 */
	public void setOrigin( Point3d origin )
	{
		this.origin = origin;
	}

	/**
	 * Returns the sphere's origin.
	 * @return The sphere's origin.
	 */
	public Point3d getOrigin()
	{
		return origin;
	}

	/**
	 * Sets the sphere's radius.
	 * @param radius The radius to set.
	 */
	public void setRadius( double radius )
	{
		this.radius = radius;
	}

	/**
	 * Returns the sphere's radius.
	 * @return The sphere's radius.
	 */
	public double getRadius()
	{
		return radius;
	}

	private Point3d origin;
	private double radius;
}