import javax.vecmath.Color3f;

/**
 * A class to contain the surface properties of an object.
 */
public class SurfaceProperties implements CloneableEx
{
	public SurfaceProperties()
	{
		setDiffuse( new Color3f() );
		setAttenuation( new Color3f() );
		setSpecularCoefficient( 0.0 );
		setSpecularExponent( 0.0 );
		setElectricPermittivity( 0.0 );
		setMagneticPermeability( 0.0 );
		setIndexOfRefraction( 0.0 );
	}
	
	/**
	 * Performs a deep copy of the specified {@code SurfaceProperties} object.
	 * 
	 * @param sp
	 * 			The {@code SurfaceProperties} object to copy.
	 */
	protected SurfaceProperties( SurfaceProperties sp )
	{
		setDiffuse( (Color3f)sp.getDiffuse().clone() );
		setAttenuation( (Color3f)sp.getAttenuation().clone() );
		setSpecularCoefficient( sp.getSpecularCoefficient() );
		setSpecularExponent( sp.getSpecularExponent() );
		setElectricPermittivity( sp.getElectricPermittivity() );
		setMagneticPermeability( sp.getMagneticPermeability() );
		setIndexOfRefraction( sp.getIndexOfRefraction() );
	}

	/**
	 * Constructs a {@code SurfaceProperties} object with the specified parameters.
	 * 
	 * @param diff
	 * 			The diffuse color of the object.
	 * @param att
	 * 			The attenuation of a ray caused by the object.
	 * @param specCoeff
	 * 			The specular coefficient of the object.
	 * @param specExp
	 * 			The specular exponent of the object.
	 * @param elecPerm
	 * 			The electric permittivitty of the object.
	 * @param magPerm
	 * 			The magnetic permeability of the object.
	 */
	public SurfaceProperties( Color3f diff, Color3f att, double specCoeff,
			double specExp, double elecPerm, double magPerm )
	{
		setDiffuse( diff );
		setAttenuation( att );
		setSpecularCoefficient( specCoeff );
		setSpecularExponent( specExp );
		setElectricPermittivity( elecPerm );
		setMagneticPermeability( magPerm );
		calculateIndexOfRefraction();
	}

	@Override
	public Object clone()
	{
		return new SurfaceProperties( this );
	}

	/**
	 * Calculates the index of refraction based on the electric permittivity
	 * and magnetic permeability.
	 */
	private void calculateIndexOfRefraction()
	{
		setIndexOfRefraction( Math.sqrt( getElectricPermittivity() * getMagneticPermeability() ) );
	}

	/**
	 * Sets the diffuse color.
	 * @param diffuse The diffuse color to set.
	 */
	public void setDiffuse( Color3f diffuse )
	{
		this.diffuse = diffuse;
	}

	/**
	 * Returns the diffuse color.
	 * @return The diffuse color.
	 */
	public Color3f getDiffuse()
	{
		return diffuse;
	}

	/**
	 * Sets the attenuation.
	 * @param attenuation The attenuation to set.
	 */
	public void setAttenuation( Color3f attenuation )
	{
		this.attenuation = attenuation;
	}

	/**
	 * Returns the attenuation.
	 * @return The attenuation.
	 */
	public Color3f getAttenuation()
	{
		return attenuation;
	}

	/**
	 * Sets the specular coefficient.
	 * @param specularCoefficient The specular coefficient to set.
	 */
	public void setSpecularCoefficient( double specularCoefficient )
	{
		this.specularCoefficient = specularCoefficient;
	}

	/**
	 * Returns the specular coefficient.
	 * @return The specular coefficient.
	 */
	public double getSpecularCoefficient()
	{
		return specularCoefficient;
	}

	/**
	 * Sets the specular exponent.
	 * @param specularExponent The specular exponent to set.
	 */
	public void setSpecularExponent( double specularExponent )
	{
		this.specularExponent = specularExponent;
	}

	/**
	 * Returns the specular exponent.
	 * @return The specular exponent.
	 */
	public double getSpecularExponent()
	{
		return specularExponent;
	}

	/**
	 * Sets the electric permittivity.
	 * @param electricPermittivity The electric permittivity to set.
	 */
	public void setElectricPermittivity( double electricPermittivity )
	{
		this.electricPermittivity = electricPermittivity;
		calculateIndexOfRefraction();
	}

	/**
	 * Returns the electric permittivity.
	 * @return The electric permittivity.
	 */
	public double getElectricPermittivity()
	{
		return electricPermittivity;
	}

	/**
	 * Sets the magnetic permeability.
	 * @param magneticPermeability The magnetic permeability to set.
	 */
	public void setMagneticPermeability( double magneticPermeability )
	{
		this.magneticPermeability = magneticPermeability;
		calculateIndexOfRefraction();
	}

	/**
	 * Returns the magnetic permeability.
	 * @return The magnetic permeability.
	 */
	public double getMagneticPermeability()
	{
		return magneticPermeability;
	}

	/**
	 * Sets the index of refraction.
	 * @param indexOfRefraction The index of refraction to set.
	 */
	private void setIndexOfRefraction( double indexOfRefraction )
	{
		this.indexOfRefraction = indexOfRefraction;
	}

	/**
	 * Returns the index of refraction as defined by the electric permittivity
	 * and the magnetic permeability.
	 * @return The index of refraction.
	 */
	public double getIndexOfRefraction()
	{
		return indexOfRefraction;
	}

	private Color3f diffuse;
	private Color3f attenuation;
	private double specularCoefficient;
	private double specularExponent;
	private double electricPermittivity;
	private double magneticPermeability;
	private double indexOfRefraction;
}
